import re
import enum

from six import iteritems


"""
task specification for JEDI

"""
class JediTaskSpec(object):
    # attributes
    attributes = (
        'jediTaskID','taskName','status','userName',
        'creationDate','modificationTime','startTime','endTime',
        'frozenTime','prodSourceLabel','workingGroup','vo','coreCount',
        'taskType','processingType','taskPriority','currentPriority',
        'architecture','transUses','transHome','transPath','lockedBy',
        'lockedTime','termCondition','splitRule','walltime','walltimeUnit',
        'outDiskCount','outDiskUnit','workDiskCount','workDiskUnit',
        'ramCount','ramUnit','ioIntensity','ioIntensityUnit',
        'workQueue_ID','progress','failureRate','errorDialog',
        'reqID','oldStatus','cloud','site','countryGroup','parent_tid',
        'eventService','ticketID','ticketSystemType','stateChangeTime',
        'superStatus','campaign','mergeRamCount','mergeRamUnit',
        'mergeWalltime','mergeWalltimeUnit','throttledTime','numThrottled',
        'mergeCoreCount','goal','assessmentTime','cpuTime','cpuTimeUnit',
        'cpuEfficiency','baseWalltime','nucleus','baseRamCount',
        'ttcRequested', 'ttcPredicted', 'ttcPredictionDate','rescueTime',
        'requestType', 'gshare', 'resource_type', 'useJumbo','diskIO',
        'diskIOUnit', 'memory_leak_core', 'memory_leak_x2', 'attemptNr',
        'container_name')
    # attributes which have 0 by default
    _zeroAttrs = ()
    # attributes to force update
    _forceUpdateAttrs = ('lockedBy','lockedTime')
    # mapping between sequence and attr
    _seqAttrMap = {}
    # limit length
    _limitLength = {'errorDialog' : 255}
    # attribute length
    _attrLength = {'workingGroup': 32}
    # tokens for split rule
    splitRuleToken = {
        'allowEmptyInput'    : 'AE',
        'addNthFieldToLFN'   : 'AN',
        'allowPartialFinish' : 'AP',
        'altStageOut'        : 'AT',
        'avoidVP'            : 'AV',
        'maxCoreCount'       : 'CC',
        'cloudAsVO'          : 'CV',
        'ddmBackEnd'         : 'DE',
        'disableAutoFinish'  : 'DF',
        'disableReassign'    : 'DI',
        'debugMode'          : 'DM',
        'disableAutoRetry'   : 'DR',
        'dynamicNumEvents'   : 'DY',
        'nEsConsumers'       : 'EC',
        'encJobParams'       : 'EJ',
        'nEventsPerWorker'   : 'ES',
        'firstContentsFeed'  : 'FC',
        'failGoalUnreached'  : 'FG',
        'firstEvent'         : 'FT',
        'groupBoundaryID'    : 'GB',
        'hpoWorkflow'        : 'HO',
        'instantiateTmplSite': 'IA',
        'inFilePosEvtNum'    : 'IF',
        'ipStack'            : 'IK',
        'allowInputLAN'      : 'IL',
        'ignoreMissingInDS'  : 'IM',
        'ipConnectivity'     : 'IP',
        'inputPreStaging'    : 'IS',
        'instantiateTmpl'    : 'IT',
        'allowInputWAN'      : 'IW',
        'noLoopingCheck'     : 'LC',
        'useLocalIO'         : 'LI',
        'limitedSites'       : 'LS',
        'loadXML'            : 'LX',
        'minCpuEfficiency'   : 'MC',
        'mergeEsOnOS'        : 'ME',
        'nMaxFilesPerJob'    : 'MF',
        'maxJumboPerSite'    : 'MJ',
        'maxNumJobs'         : 'MN',
        'mergeOutput'        : 'MO',
        'multiStepExec'      : 'MS',
        'maxWalltime'        : 'MW',
        'noExecStrCnv'       : 'NC',
        'notDiscardEvents'   : 'ND',
        'nEventsPerJob'      : 'NE',
        'nFilesPerJob'       : 'NF',
        'nGBPerJob'          : 'NG',
        'noInputPooling'     : 'NI',
        'nJumboJobs'         : 'NJ',
        'nSitesPerJob'       : 'NS',
        'nChunksToWait'      : 'NT',
        'noWaitParent'       : 'NW',
        'orderByLB'          : 'OL',
        'osMatching'         : 'OS',
        'onlyTagsForFC'      : 'OT',
        'pushStatusChanges'  : 'PC',
        'pfnList'            : 'PL',
        'putLogToOS'         : 'PO',
        'runUntilClosed'     : 'RC',
        'registerDatasets'   : 'RD',
        'registerEsFiles'    : 'RE',
        'respectLB'          : 'RL',
        'reuseSecOnDemand'   : 'RO',
        'releasePerLB'       : 'RP',
        'respectSplitRule'   : 'RR',
        'randomSeed'         : 'RS',
        'resurrectConsumers' : 'SC',
        'switchEStoNormal'   : 'SE',
        'stayOutputOnSite'   : 'SO',
        'scoutSuccessRate'   : 'SS',
        'useSecrets'         : 'ST',
        'segmentedWork'      : 'SW',
        'totNumJobs'         : 'TJ',
        'tgtMaxOutputForNG'  : 'TN',
        't1Weight'           : 'TW',
        'useBuild'           : 'UB',
        'useJobCloning'      : 'UC',
        'useRealNumEvents'   : 'UE',
        'useFileAsSourceLFN' : 'UF',
        'usePrePro'          : 'UP',
        'useScout'           : 'US',
        'usePrefetcher'      : 'UT',
        'useExhausted'       : 'UX',
        'useZipToPin'        : 'UZ',
        'writeInputToFile'   : 'WF',
        'waitInput'          : 'WI',
        'maxAttemptES'       : 'XA',
        'decAttOnFailedES'   : 'XF',
        'maxAttemptEsJob'    : 'XJ',
        'nEventsPerMergeJob'   : 'ZE',
        'nFilesPerMergeJob'    : 'ZF',
        'nGBPerMergeJob'       : 'ZG',
        'nMaxFilesPerMergeJob' : 'ZM',

        }
    # enum for preprocessing
    enum_toPreProcess = '1'
    enum_preProcessed = '2'
    enum_postPProcess = '3'
    # enum for limited sites
    enum_limitedSites = {'1' : 'inc',
                         '2' : 'exc',
                         '3' : 'incexc'}
    # enum for scout
    enum_noScout   = '1'
    enum_useScout  = '2'
    enum_postScout = '3'
    # enum for dataset registration
    enum_toRegisterDS   = '1'
    enum_registeredDS   = '2'
    # enum for IP connectivity
    enum_ipConnectivity = {'1' : 'full',
                           '2' : 'http'}
    # enum for IP stack
    enum_ipStack = {'4' : 'IPv4',
                    '6' : 'IPv6'}
    # enum for alternative stage-out
    enum_altStageOut = {'1' : 'on',
                        '2' : 'off',
                        '3' : 'force'}
    # enum for local direct access
    enum_inputLAN = {'1' : 'use',
                     '2' : 'only'}
    # world cloud name
    worldCloudName = 'WORLD'

    # enum for contents feeder
    class FirstContentsFeed(enum.Enum):
        TRUE = '1'
        FALSE = '0'

    # enum for useJumbo
    enum_useJumbo = {'waiting': 'W',
                     'running': 'R',
                     'pending': 'P',
                     'lack': 'L',
                     'disabled': 'D'}

    # enum for input prestaging
    enum_inputPreStaging = {'use': '1',
                            'notUse': '0'}


    # constructor
    def __init__(self):
        # install attributes
        for attr in self.attributes:
            if attr in self._zeroAttrs:
                object.__setattr__(self,attr,0)
            else:
                object.__setattr__(self,attr,None)
        # map of changed attributes
        object.__setattr__(self,'_changedAttrs',{})
        # template to generate job parameters
        object.__setattr__(self,'jobParamsTemplate','')
        # associated datasets
        object.__setattr__(self,'datasetSpecList',[])
        # original error dialog
        object.__setattr__(self, 'origErrorDialog', None)
        # original user name
        object.__setattr__(self, 'origUserName', None)


    # override __setattr__ to collect the changed attributes
    def __setattr__(self, name, value):
        oldVal = getattr(self,name)
        if name in self._limitLength and value is not None:
            # keep original dialog
            if name == 'errorDialog':
                object.__setattr__(self, 'origErrorDialog', value)
            value = value[:self._limitLength[name]]
        object.__setattr__(self,name,value)
        newVal = getattr(self,name)
        # collect changed attributes
        if oldVal != newVal or name in self._forceUpdateAttrs:
            self._changedAttrs[name] = value


    # copy old attributes
    def copyAttributes(self,oldTaskSpec):
        for attr in self.attributes + ('jobParamsTemplate',):
            if 'Time' in attr:
                continue
            if 'Date' in attr:
                continue
            if attr in ['progress','failureRate','errorDialog',
                        'status','oldStatus','lockedBy']:
                continue
            self.__setattr__(attr,getattr(oldTaskSpec,attr))


    # reset changed attribute list
    def resetChangedList(self):
        object.__setattr__(self,'_changedAttrs',{})


    # reset changed attribute
    def resetChangedAttr(self,name):
        try:
            del self._changedAttrs[name]
        except Exception:
            pass


    # force update
    def forceUpdate(self,name):
        if name in self.attributes:
            self._changedAttrs[name] = getattr(self,name)


    # return map of values
    def valuesMap(self,useSeq=False,onlyChanged=False):
        ret = {}
        for attr in self.attributes:
            # use sequence
            if useSeq and attr in self._seqAttrMap:
                continue
            # only changed attributes
            if onlyChanged:
                if attr not in self._changedAttrs:
                    continue
            val = getattr(self,attr)
            if val is None:
                if attr in self._zeroAttrs:
                    val = 0
                else:
                    val = None
            # truncate too long values
            if attr in self._limitLength:
                if val is not None:
                    val = val[:self._limitLength[attr]]
            ret[':%s' % attr] = val
        return ret


    # pack tuple into TaskSpec
    def pack(self,values):
        for i in range(len(self.attributes)):
            attr= self.attributes[i]
            val = values[i]
            object.__setattr__(self,attr,val)


    # return column names for INSERT
    def columnNames(cls,prefix=None):
        ret = ""
        for attr in cls.attributes:
            if prefix is not None:
                ret += '{0}.'.format(prefix)
            ret += '{0},'.format(attr)
        ret = ret[:-1]
        return ret
    columnNames = classmethod(columnNames)


    # return expression of bind variables for INSERT
    def bindValuesExpression(cls,useSeq=True):
        ret = "VALUES("
        for attr in cls.attributes:
            if useSeq and attr in cls._seqAttrMap:
                ret += "%s," % cls._seqAttrMap[attr]
            else:
                ret += ":%s," % attr
        ret = ret[:-1]
        ret += ")"
        return ret
    bindValuesExpression = classmethod(bindValuesExpression)


    # return an expression of bind variables for UPDATE to update only changed attributes
    def bindUpdateChangesExpression(self):
        ret = ""
        for attr in self.attributes:
            if attr in self._changedAttrs:
                ret += '%s=:%s,' % (attr,attr)
        ret  = ret[:-1]
        ret += ' '
        return ret



    # get the max size per job if defined
    def getMaxSizePerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nGBPerJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                nGBPerJob = int(tmpMatch.group(1)) * 1024 * 1024 * 1024
                return nGBPerJob
        return None

    # remove nGBPerJob
    def removeMaxSizePerJob(self):
        self.removeSplitRule(self.splitRuleToken['nGBPerJob'])

    # get the max size per merge job if defined
    def getMaxSizePerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nGBPerMergeJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                nGBPerJob = int(tmpMatch.group(1)) * 1024 * 1024 * 1024
                return nGBPerJob
        return None


    # get the maxnumber of files per job if defined
    def getMaxNumFilesPerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nMaxFilesPerJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # set MaxNumFilesPerJob
    def setMaxNumFilesPerJob(self, value):
        self.setSplitRule('nMaxFilesPerJob', value)

    # set


    # get the maxnumber of files per merge job if defined
    def getMaxNumFilesPerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nMaxFilesPerMergeJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 50



    # get the number of events per merge job if defined
    def getNumEventsPerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nEventsPerMergeJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # check if using jumbo
    def usingJumboJobs(self):
        if self.useJumbo in self.enum_useJumbo.values() and self.useJumbo != self.enum_useJumbo['disabled']:
            return True
        return False



    # get the number of jumbo jobs if defined
    def getNumJumboJobs(self):
        if self.usingJumboJobs() and self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nJumboJobs']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None




    # get the max number of jumbo jobs per site if defined
    def getMaxJumboPerSite(self):
        if self.usingJumboJobs() and self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['maxJumboPerSite']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 1



    # get the number of sites per job
    def getNumSitesPerJob(self):
        if not self.useEventService():
            return 1
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nSitesPerJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 1



    # get the number of files per job if defined
    def getNumFilesPerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nFilesPerJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # remove nFilesPerJob
    def removeNumFilesPerJob(self):
        self.removeSplitRule(self.splitRuleToken['nFilesPerJob'])

    # get the number of files per merge job if defined
    def getNumFilesPerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nFilesPerMergeJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # get the number of events per job if defined
    def getNumEventsPerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nEventsPerJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # get offset for random seed
    def getRndmSeedOffset(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['randomSeed']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 0



    # get offset for first event
    def getFirstEventOffset(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['firstEvent']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 0



    # grouping with boundaryID
    def useGroupWithBoundaryID(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['groupBoundaryID']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                gbID = int(tmpMatch.group(1))
                # 1 : input - can split,    output - free
                # 2 : input - can split,    output - mapped with provenanceID
                # 3 : input - cannot split, output - free
                # 4 : input - cannot split, output - mapped with provenanceID
                #
                # * rule for master
                # 1 : can split. one boundayID per sub chunk
                # 2 : cannot split. one boundayID per sub chunk
                # 3 : cannot split. multiple boundayIDs per sub chunk
                #
                # * rule for secodary
                # 1 : must have same boundayID. cannot split
                #
                retMap = {}
                if gbID in [1,2]:
                    retMap['inSplit'] = 1
                else:
                    retMap['inSplit'] = 2
                if gbID in [1,3]:
                    retMap['outMap'] = False
                else:
                    retMap['outMap'] = True
                retMap['secSplit'] = None
                return retMap
        return None



    # use build
    def useBuild(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useBuild']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use sjob cloning
    def useJobCloning(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useJobCloning']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # get job cloning type
    def getJobCloningType(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useJobCloning']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return tmpMatch.group(1)
        return ''



    # reuse secondary on demand
    def reuseSecOnDemand(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['reuseSecOnDemand']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # not wait for completion of parent
    def noWaitParent(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['noWaitParent']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # check splitRule if not wait for completion of parent
    def noWaitParentSL(cls,splitRule):
        if splitRule is not None:
            tmpMatch = re.search(cls.splitRuleToken['noWaitParent']+'=(\d+)',splitRule)
            if tmpMatch is not None:
                return True
        return False
    noWaitParentSL = classmethod(noWaitParentSL)



    # use only limited sites
    def useLimitedSites(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['limitedSites']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # set limited sites
    def setLimitedSites(self,policy):
        tag = None
        for tmpIdx,tmpPolicy in iteritems(self.enum_limitedSites):
            if policy == tmpPolicy:
                tag = tmpIdx
                break
        # not found
        if tag is None:
            return
        # set
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken['limitedSites']+'='+tag
        else:
            tmpMatch = re.search(self.splitRuleToken['limitedSites']+'=(\d+)',self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += ','+self.splitRuleToken['limitedSites']+'='+tag
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken['limitedSites']+'=(\d+)',
                                        self.splitRuleToken['limitedSites']+'='+tag,
                                        self.splitRule)



    # use local IO
    def useLocalIO(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useLocalIO']+'=(\d+)',self.splitRule)
            if tmpMatch is not None and int(tmpMatch.group(1)):
                return True
        return False



    # use Event Service
    def useEventService(self,siteSpec=None):
        if self.eventService in [1,2]:
            # check site if ES is disabled
            if self.switchEStoNormal() and siteSpec is not None and siteSpec.getJobSeed() in ['all']:
                return False
            return True
        return False



    # get the number of events per worker for Event Service
    def getNumEventsPerWorker(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nEventsPerWorker']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None




    # get the number of event service consumers
    def getNumEventServiceConsumer(self):
        if not self.useEventService():
            return None
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nEsConsumers']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # disable automatic retry
    def disableAutoRetry(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['disableAutoRetry']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # disable reassign
    def disableReassign(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['disableReassign']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # allow empty input
    def allowEmptyInput(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['allowEmptyInput']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use PFN list
    def useListPFN(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['pfnList']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # set preprocessing
    def setPrePro(self):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken['usePrePro']+'='+self.enum_toPreProcess
        else:
            # append
            self.splitRule += ','+self.splitRuleToken['usePrePro']+'='+self.enum_toPreProcess



    # use preprocessing
    def usePrePro(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['usePrePro']+'=(\d+)',self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_toPreProcess:
                return True
        return False



    # set preprocessed
    def setPreProcessed(self):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken['usePrePro']+'='+self.enum_preProcessed
        else:
            tmpMatch = re.search(self.splitRuleToken['usePrePro']+'=(\d+)',self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += ','+self.splitRuleToken['usePrePro']+'='+self.enum_preProcessed
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken['usePrePro']+'=(\d+)',
                                        self.splitRuleToken['usePrePro']+'='+self.enum_preProcessed,
                                        self.splitRule)
        return



    # check preprocessed
    def checkPreProcessed(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['usePrePro']+'=(\d+)',self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_preProcessed:
                return True
        return False



    # set post preprocess
    def setPostPreProcess(self):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken['usePrePro']+'='+self.enum_postPProcess
        else:
            tmpMatch = re.search(self.splitRuleToken['usePrePro']+'=(\d+)',self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += ','+self.splitRuleToken['usePrePro']+'='+self.enum_postPProcess
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken['usePrePro']+'=(\d+)',
                                        self.splitRuleToken['usePrePro']+'='+self.enum_postPProcess,
                                        self.splitRule)
        return



    # instantiate template datasets
    def instantiateTmpl(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['instantiateTmpl']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # instantiate template datasets at site
    def instantiateTmplSite(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['instantiateTmplSite']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # merge output
    def mergeOutput(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['mergeOutput']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use random seed
    def useRandomSeed(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['randomSeed']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # get the size of workDisk in bytes
    def getWorkDiskSize(self):
        safetyMargin = 300 * 1024 * 1024
        tmpSize = self.workDiskCount
        if tmpSize is None:
            return 0
        if self.workDiskUnit == 'GB':
            tmpSize = tmpSize * 1024 * 1024 * 1024
        elif self.workDiskUnit == 'MB':
            tmpSize = tmpSize * 1024 * 1024
        elif self.workDiskUnit == 'kB':
            tmpSize = tmpSize * 1024
        tmpSize += safetyMargin
        return tmpSize



    # get the size of outDisk in bytes
    def getOutDiskSize(self):
        tmpSize = self.outDiskCount
        if tmpSize is None or tmpSize < 0:
            return 0
        if self.outDiskUnit is not None:
            if self.outDiskUnit.startswith('GB'):
                tmpSize = tmpSize * 1024 * 1024 * 1024
            elif self.outDiskUnit.startswith('MB'):
                tmpSize = tmpSize * 1024 * 1024
            elif self.outDiskUnit.startswith('kB'):
                tmpSize = tmpSize * 1024
        return tmpSize



    # output scales with the number of events
    def outputScaleWithEvents(self):
        if self.outDiskUnit is not None and 'PerEvent' in self.outDiskUnit:
            return True
        return False



    # return list of status to update contents
    def statusToUpdateContents(cls):
        return ['defined']
    statusToUpdateContents = classmethod(statusToUpdateContents)



    # set task status on hold
    def setOnHold(self):
        # change status
        if self.status in ['ready','running','merging','scouting','defined',
                           'topreprocess','preprocessing','registered',
                           'prepared','rerefine']:
            self.oldStatus = self.status
            self.status = 'pending'


    # return list of status to reject external changes
    def statusToRejectExtChange(cls):
        return ['finished','done','prepared','broken','tobroken','aborted','toabort','aborting','failed',
                'finishing','passed']
    statusToRejectExtChange = classmethod(statusToRejectExtChange)



    # return list of status for retry
    def statusToRetry(cls):
        return ['finished','failed','aborted','exhausted']
    statusToRetry = classmethod(statusToRetry)



    # return list of status for incexec
    def statusToIncexec(cls):
        return ['done'] + cls.statusToRetry()
    statusToIncexec = classmethod(statusToIncexec)



    # return list of status for reassign
    def statusToReassign(cls):
        return ['registered','defined','ready','running','scouting','scouted','pending',
                'assigning','exhausted']
    statusToReassign = classmethod(statusToReassign)



    # return list of status for Job Generator
    def statusForJobGenerator(cls):
        return ['ready','running','scouting','topreprocess','preprocessing']
    statusForJobGenerator = classmethod(statusForJobGenerator)



    # return list of status to not pause
    def statusNotToPause(cls):
        return ['finished','failed','done','aborted','broken','paused']
    statusNotToPause = classmethod(statusNotToPause)



    # return mapping of command and status
    def commandStatusMap(cls):
        return {'kill' : {'doing': 'aborting',
                          'done' : 'toabort'},
                'finish' : {'doing': 'finishing',
                            'done' : 'passed'},
                'retry' : {'doing': 'toretry',
                           'done' : 'ready'},
                'incexec' : {'doing': 'toincexec',
                             'done' : 'rerefine'},
                'reassign' : {'doing': 'toreassign',
                              'done' : 'reassigning'},
                'pause' : {'doing': 'paused',
                           'done' : 'dummy'},
                'resume' : {'doing': 'dummy',
                            'done' : 'dummy'},
                'avalanche' : {'doing': 'dummy',
                               'done' : 'dummy'},
                }
    commandStatusMap = classmethod(commandStatusMap)



    # set task status to active
    def setInActive(self):
        # change status
        if self.status == 'pending':
            self.status = self.oldStatus
            self.oldStatus = None
            self.errorDialog = None
        # reset error dialog
        self.setErrDiag(None)


    # set error dialog
    def setErrDiag(self,diag,append=False,prepend=False):
        # set error dialog
        if append is True and self.errorDialog is not None:
            if not prepend:
                self.errorDialog = "{0} {1}".format(self.errorDialog,diag)
            else:
                self.errorDialog = "{0} {1}".format(diag,self.errorDialog)
        elif append is None:
            # keep old log
            if self.errorDialog is None:
                self.errorDialog = diag
        else:
            self.errorDialog = diag



    # use loadXML
    def useLoadXML(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['loadXML']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # make VOMS FQANs
    def makeFQANs(self):
        # no working group
        if self.workingGroup is not None:
            fqan = '/{0}/{1}/Role=production'.format(self.vo,self.workingGroup)
        else:
            if self.vo is not None:
                fqan = '/{0}/Role=NULL'.format(self.vo)
            else:
                return []
        # return
        return [fqan]



    # set split rule
    def setSplitRule(self,ruleName,ruleValue):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken[ruleName]+'='+ruleValue
        else:
            tmpMatch = re.search(self.splitRuleToken[ruleName]+'=(\d+)',self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += ','+self.splitRuleToken[ruleName]+'='+ruleValue
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken[ruleName]+'=(\d+)',
                                        self.splitRuleToken[ruleName]+'='+ruleValue,
                                        self.splitRule)



    # remove split rule
    def removeSplitRule(self,ruleName):
        if self.splitRule is not None:
            items = self.splitRule.split(',')
            newItems = []
            for item in items:
                # remove rile
                tmpRuleName = item.split('=')[0]
                if ruleName == tmpRuleName:
                    continue
                newItems.append(item)
            self.splitRule = ','.join(newItems)



    # set to use scout
    def setUseScout(self,useFlag):
        if useFlag:
            self.setSplitRule('useScout',self.enum_useScout)
        else:
            self.setSplitRule('useScout',self.enum_noScout)



    # set post scout
    def setPostScout(self):
        self.setSplitRule('useScout',self.enum_postScout)



    # use scout
    def useScout(self,splitRule=None):
        if splitRule is None:
            splitRule = self.splitRule
        if splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useScout']+'=(\d+)',splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_useScout:
                return True
        return False



    # use exhausted
    def useExhausted(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useExhausted']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use real number of events
    def useRealNumEvents(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useRealNumEvents']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use input LFN as source for output LFN
    def useFileAsSourceLFN(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useFileAsSourceLFN']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # post scout
    def isPostScout(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useScout']+'=(\d+)',self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_postScout:
                return True
        return False



    # wait until input shows up
    def waitInput(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['waitInput']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False


    # input prestaging
    def inputPreStaging(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['inputPreStaging']+'='+self.enum_inputPreStaging['use'],
                                 self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # disable input prestaging
    def disableInputPreStaging(self):
        self.setSplitRule('inputPreStaging', self.enum_inputPreStaging['notUse'])

    # set DDM backend
    def setDdmBackEnd(self,backEnd):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken['ddmBackEnd']+'='+backEnd
        else:
            tmpMatch = re.search(self.splitRuleToken['ddmBackEnd']+'=([^,$]+)',self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += ','+self.splitRuleToken['ddmBackEnd']+'='+backEnd
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken['ddmBackEnd']+'=([^,$]+)',
                                        self.splitRuleToken['ddmBackEnd']+'='+backEnd,
                                        self.splitRule)



    # get DDM backend
    def getDdmBackEnd(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['ddmBackEnd']+'=([^,$]+)',self.splitRule)
            if tmpMatch is not None:
                return tmpMatch.group(1)
        return None



    # get field number to add middle name to LFN
    def getFieldNumToLFN(self):
        try:
            if self.splitRule is not None:
                tmpMatch = re.search(self.splitRuleToken['addNthFieldToLFN']+'=([,\d]+)',self.splitRule)
                if tmpMatch is not None:
                    tmpList = tmpMatch.group(1).split(',')
                    try:
                        tmpList.remove('')
                    except Exception:
                        pass
                    return list(map(int, tmpList))
        except Exception:
            pass
        return None



    # get required success rate for scout jobs
    def getScoutSuccessRate(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['scoutSuccessRate']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # get T1 weight
    def getT1Weight(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['t1Weight']+'=(-*\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 0



    # respect Lumiblock boundaries
    def respectLumiblock(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['respectLB']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # release files per Lumiblock
    def releasePerLumiblock(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['releasePerLB']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # order by Lumiblock numbers
    def orderByLB(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['orderByLB']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # respect split rule
    def respectSplitRule(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['respectSplitRule']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # allow partial finish
    def allowPartialFinish(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['allowPartialFinish']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # check if datasets should be registered
    def toRegisterDatasets(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['registerDatasets']+'=(\d+)',self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_toRegisterDS:
                return True
        return False



    # datasets were registered
    def registeredDatasets(self):
        self.setSplitRule('registerDatasets',self.enum_registeredDS)



    # set datasets to be registered
    def setToRegisterDatasets(self):
        self.setSplitRule('registerDatasets',self.enum_toRegisterDS)



    # get the max number of attempts for ES events
    def getMaxAttemptES(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['maxAttemptES']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # get the max number of attempts for ES jobs
    def getMaxAttemptEsJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['maxAttemptEsJob']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return self.getMaxAttemptES()



    # check attribute length
    def checkAttrLength(self):
        for attrName,attrLength in iteritems(self._attrLength):
            attrVal = getattr(self,attrName)
            if attrVal is None:
                continue
            if len(attrVal) > attrLength:
                setattr(self,attrName,None)
                self.errorDialog = "{0} is too long (actual: {1}, maximum: {2})".format(attrName,len(attrVal),attrLength)
                return False
        return True

    # set IP connectivity and stack
    def setIpConnectivity(self, value):
        if not value:
            return
        values = value.split('#')
        if not values:
            return
        ipConnectivity = values[0]
        if ipConnectivity in self.enum_ipConnectivity.values():
            for tmpKey, tmpVal in iteritems(self.enum_ipConnectivity):
                if ipConnectivity == tmpVal:
                    self.setSplitRule('ipConnectivity', tmpKey)
                    break
        if len(values) > 1:
            ipStack = values[1]
            if ipStack in self.enum_ipStack.values():
                for tmpKey, tmpVal in iteritems(self.enum_ipStack):
                    if ipStack == tmpVal:
                        self.setSplitRule('ipStack', tmpKey)
                        break

    # get IP connectivity
    def getIpConnectivity(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['ipConnectivity']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return self.enum_ipConnectivity[tmpMatch.group(1)]
        return None

    # get IP connectivity
    def getIpStack(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['ipStack']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return self.enum_ipStack[tmpMatch.group(1)]
        return None

    # use HS06 for walltime estimation
    def useHS06(self):
        return self.cpuTimeUnit in ['HS06sPerEvent', 'HS06sPerEventFixed', 'mHS06sPerEvent', 'mHS06sPerEventFixed']

    # get CPU time in sec
    def getCpuTime(self):
        if not self.useHS06():
            return None
        try:
            if self.cpuTimeUnit.startswith('m'):
                return float(self.cpuTime) / 1000.0
        except Exception:
            pass
        return self.cpuTime

    # RAM scales with nCores
    def ramPerCore(self):
        return self.ramUnit in ['MBPerCore','MBPerCoreFixed']



    # run until input is closed
    def runUntilClosed(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['runUntilClosed']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # stay output on site
    def stayOutputOnSite(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['stayOutputOnSite']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # fail when goal unreached
    def failGoalUnreached(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['failGoalUnreached']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # unset fail when goal unreached
    def unsetFailGoalUnreached(self):
        self.removeSplitRule(self.splitRuleToken['failGoalUnreached'])



    # switch ES to normal when jobs land at normal sites
    def switchEStoNormal(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['switchEStoNormal']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use world cloud
    def useWorldCloud(self):
        return self.cloud == self.worldCloudName



    # dynamic number of events
    def dynamicNumEvents(self):
        if self.getNumEventsPerJob() is not None:
            return False
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['dynamicNumEvents']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # get min granularity for dynamic number of events
    def get_min_granularity(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['dynamicNumEvents']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # set alternative stage-out
    def setAltStageOut(self,value):
        if value in self.enum_altStageOut.values():
            for tmpKey,tmpVal in iteritems(self.enum_altStageOut):
                if value == tmpVal:
                    self.setSplitRule('altStageOut',tmpKey)
                    break



    # get alternative stage-out
    def getAltStageOut(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['altStageOut']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return self.enum_altStageOut[tmpMatch.group(1)]
        return None



    # allow WAN for input access
    def allowInputWAN(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['allowInputWAN']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # set mode for input LAN access
    def setAllowInputLAN(self,value):
        if value in self.enum_inputLAN.values():
            for tmpKey,tmpVal in iteritems(self.enum_inputLAN):
                if value == tmpVal:
                    self.setSplitRule('allowInputLAN',tmpKey)
                    break



    # check if LAN is used for input access
    def allowInputLAN(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['allowInputLAN']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return self.enum_inputLAN[tmpMatch.group(1)]
        return None



    # put log files to OS
    def putLogToOS(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['putLogToOS']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # merge ES on Object Store
    def mergeEsOnOS(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['mergeEsOnOS']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # write input to file
    def writeInputToFile(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['writeInputToFile']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # ignore missing input datasets
    def ignoreMissingInDS(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['ignoreMissingInDS']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # suppress execute string conversion
    def noExecStrCnv(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['noExecStrCnv']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # in-file positional event number
    def inFilePosEvtNum(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['inFilePosEvtNum']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # register event service files
    def registerEsFiles(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['registerEsFiles']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # disable auto finish
    def disableAutoFinish(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['disableAutoFinish']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # reset refined attributes which may confuse they system
    def resetRefinedAttrs(self):
        self.resetChangedAttr('splitRule')
        self.resetChangedAttr('eventService')



    # resurrect consumers
    def resurrectConsumers(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['resurrectConsumers']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use prefetcher
    def usePrefetcher(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['usePrefetcher']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # no input pooling
    def noInputPooling(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['noInputPooling']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # get num of input chunks to wait
    def nChunksToWait(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['nChunksToWait']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None


    # get max walltime
    def getMaxWalltime(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['maxWalltime']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1)) * 60 * 60
        return None



    # get target size of the largest output to reset NG
    def getTgtMaxOutputForNG(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['tgtMaxOutputForNG']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # not discard events
    def notDiscardEvents(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['notDiscardEvents']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # get min CPU efficiency
    def getMinCpuEfficiency(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['minCpuEfficiency']+'=(\d+)',self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None



    # decrement attemptNr of events only when failed
    def decAttOnFailedES(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['decAttOnFailedES']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # use zip files to pin input files
    def useZipToPin(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['useZipToPin']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False



    # require OS matching
    def osMatching(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['osMatching']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # architecture: sw_platform<@base_platform><#host_cpu_spec><&host_gpu_spec>
    # host_cpu_spec: architecture<-vendor<-instruction_set>>
    # host_gpu_spec: vendor<-model>

    # get SW platform
    def get_sw_platform(self):
        if self.architecture is not None:
            m = re.search('^([^@#&]*)', self.architecture)
            if m:
                return m.group(1)
        return self.architecture

    # get base platform
    def get_base_platform(self):
        if self.architecture is None or '@' not in self.architecture:
            return None
        m = re.search('@([^#&]*)', self.architecture)
        img = m.group(1)
        if img == '':
            img = None
        return img

    # get platforms
    def get_platforms(self):
        if self.architecture is not None:
            platform = self.get_sw_platform()
            base = self.get_base_platform()
            if base:
                platform += '@' + base
            return  platform
        return self.architecture

    # get host CPU spec
    def get_host_cpu_spec(self):
        try:
            if self.architecture is None or '#' not in self.architecture:
                return None
            m = re.search('#([^@&]*)', self.architecture)
            spec_str = m.group(1)
            if not spec_str:
                return None
            spec_str += '-*' * (2 - spec_str.count('-'))
            if '-' not in spec_str:
                spec_str += '-*'
            items = spec_str.split('-')
            spec = {'arch': items[0],
                    'vendor': items[1],
                    'instr': items[2]}
            return spec
        except Exception:
            return None

    # get host GPU spec
    def get_host_gpu_spec(self):
        try:
            if self.architecture is None or '&' not in self.architecture:
                return None
            m = re.search('&([^@#]*)', self.architecture)
            spec_str = m.group(1)
            if not spec_str:
                return None
            spec_str += '-*' * (1 - spec_str.count('-'))
            items = spec_str.split('-')
            spec = {'vendor': items[0],
                    'model': items[1]}
            return spec
        except Exception:
            return None

    # HPO workflow
    def is_hpo_workflow(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['hpoWorkflow']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # debug mode
    def is_debug_mode(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['debugMode']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # multi-step execution
    def is_multi_step_exec(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['multiStepExec']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # get max number of jobs
    def get_max_num_jobs(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['maxNumJobs']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get total number of jobs
    def get_total_num_jobs(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['totNumJobs']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # use only tags for fat container
    def use_only_tags_fc(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['onlyTagsForFC']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # avoid VP
    def avoid_vp(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['avoidVP']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # set first contents feed
    def set_first_contents_feed(self, is_first):
        if is_first:
            self.setSplitRule('firstContentsFeed', self.FirstContentsFeed.TRUE.value)
        else:
            self.setSplitRule('firstContentsFeed', self.FirstContentsFeed.FALSE.value)

    # check if first contents feed
    def is_first_contents_feed(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['firstContentsFeed']+'=(\d+)', self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.FirstContentsFeed.TRUE.value:
                return True
        return False

    # check if work is segmented
    def is_work_segmented(self):
        if self.splitRule is not None:
            if re.search(self.splitRuleToken['segmentedWork']+'=(\d+)', self.splitRule):
                return True
        return False

    # check if looping check is disabled
    def no_looping_check(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['noLoopingCheck']+'=(\d+)', self.splitRule)
            if tmpMatch and int(tmpMatch.group(1)):
                return True
        return False

    # encode job parameters
    def encode_job_params(self):
        if self.splitRule is not None:
            if re.search(self.splitRuleToken['encJobParams']+'=(\d+)', self.splitRule):
                return True
        return False

    # get original error dialog
    def get_original_error_dialog(self):
        if not self.origErrorDialog:
            return ''
        # remove log URL
        tmpStr = re.sub('<a href.+</a> : ', '', self.origErrorDialog)
        return tmpStr.split('. ')[-1]

    # check if secrets are used
    def use_secrets(self):
        if self.splitRule is not None:
            if re.search(self.splitRuleToken['useSecrets']+'=(\d+)', self.splitRule):
                return True
        return False

    # get max core count
    def get_max_core_count(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken['maxCoreCount']+'=(\d+)', self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # push status changes
    def push_status_changes(self):
        return push_status_changes(self.splitRule)

    # use cloud as VO
    def cloud_as_vo(self):
        if self.splitRule is not None:
            if re.search(self.splitRuleToken['cloudAsVO']+'=(\d+)', self.splitRule):
                return True
        return False


# utils

# check if push status changes without class instance
def push_status_changes(split_rule):
    if not split_rule:
        return False
    tmpMatch = re.search(JediTaskSpec.splitRuleToken['pushStatusChanges'] + r'=(\d+)', split_rule)
    if not tmpMatch or int(tmpMatch.group(1)) <= 0:
        return False
    return True



from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface

# Set up the task buffer interface
tb_if = JediTaskBufferInterface()
tb_if.setupInterface()

# Get the site mapper
site_mapper = tb_if.getSiteMapper()

# Define a random site of panda sites
panda_sites = ['ANALY_DESY-HH', 'ANALY_DESY-HH_TEST', 'WT2_Install', 'BNL_Test_2_CE_1', 'ANALY_IN2P3-CC-T2_RD',
               'UKI-SOUTHGRID-OX-HEP_SL6', 'RU-Protvino-IHEP', 'ANALY_NICS_Kraken', 'ANALY_TRIUMF_HIMEM',
               'TESTGLEXEC', 'ANALY_GRIF-LPNHE', 'UKI-SCOTGRID-GLASGOW_MCORE', 'BNL_PROD', 'RAL-LCG2_SL6']

# Test the getSiteInputStorageEndpointMap function
# print tb_if.getSiteInputStorageEndpointMap(panda_sites, site_mapper)

# for panda_site in panda_sites:
#     print panda_site
#     tmp_site_spec = site_mapper.getSite('ANALY_DESY-HH')
#     print '------------------- ddm -------------------'
#     print 'ddm_input: {0}, ddm_output: {1}'.format(tmp_site_spec.ddm_input, tmp_site_spec.ddm_output)
#     print '------------------- setokens values -------------------'
#     print 'setokens_input: {0}, setokens_output: {1}'.format(tmp_site_spec.setokens_input.values(),
#                                                              tmp_site_spec.setokens_output.values())
#     print '------------------- setokens -------------------'
#     print 'setokens_input: {0}, setokens_output: {1}'.format(tmp_site_spec.setokens_input,
#                                                              tmp_site_spec.setokens_output)

from pandaserver.dataservice import DataServiceUtils    # noqa: E402
sites = site_mapper.getCloud('WORLD')['sites']
sites.sort()
for tmp_site_name in sites:
    print('tmp_site_name: {0}'.format(tmp_site_name))

    tmp_site_spec = site_mapper.getSite(tmp_site_name)

    #print 'tmp_site_spec.ddm_input: {0}'.format(tmp_site_spec.ddm_input)
    #print 'tmp_site_spec.setokens_input: {0}'.format(tmp_site_spec.setokens_input.values())
    #print 'combination: {0}'.format([tmp_site_spec.ddm_input] + tmp_site_spec.setokens_input.values())

    for tmp_ddm_endpoint in [tmp_site_spec.ddm_input] + list(tmp_site_spec.setokens_input.values()):
        try:
            tmp_prefix = DataServiceUtils.getDQ2Prefix(tmp_ddm_endpoint)
            print('prefix: {0}'.format(tmp_prefix))
        except TypeError:
            print('excepted!')

    print('-------------------')



import re
import json

import six
from six import iteritems


# convert UTF-8 to ASCII in json dumps
def unicodeConvert(input):
    if isinstance(input, dict):
        retMap = {}
        for tmpKey,tmpVal in iteritems(input):
            retMap[unicodeConvert(tmpKey)] = unicodeConvert(tmpVal)
        return retMap
    elif isinstance(input, list):
        retList = []
        for tmpItem in input:
            retList.append(unicodeConvert(tmpItem))
        return retList
    elif six.PY2 and isinstance(input, unicode):
        return input.encode('ascii', 'ignore')
    return input



# decode
def decodeJSON(inputStr):
    return json.loads(inputStr, object_hook=unicodeConvert)



# encode
def encodeJSON(inputMap):
    return json.dumps(inputMap)



# extract stream name
def extractStreamName(valStr):
    tmpMatch = re.search('\$\{([^\}]+)\}',valStr)
    if tmpMatch is None:
        return None
    # remove decorators
    streamName = tmpMatch.group(1)
    streamName = streamName.split('/')[0]
    return streamName



# extract output filename template and replace the value field
def extractReplaceOutFileTemplate(valStr,streamName):
    outFileTempl = valStr.split('=')[-1]
    outFileTempl = outFileTempl.replace("'","")
    valStr = valStr.replace(outFileTempl,'${{{0}}}'.format(streamName))
    return outFileTempl,valStr



# extract file list
def extractFileList(taskParamMap,datasetName):
    baseDatasetName = datasetName.split(':')[-1]
    if 'log' in taskParamMap:
        itemList = taskParamMap['jobParameters'] + [taskParamMap['log']]
    else:
        itemList = taskParamMap['jobParameters']
    fileList = []
    includePatt = []
    excludePatt = []
    for tmpItem in itemList:
        if tmpItem['type'] == 'template' and 'dataset' in tmpItem and \
                ((tmpItem['dataset'] == datasetName or tmpItem['dataset'] == baseDatasetName) or \
                     ('expandedList' in tmpItem and \
                          (datasetName in tmpItem['expandedList'] or baseDatasetName in tmpItem['expandedList']))):
            if 'files' in tmpItem:
                fileList = tmpItem['files']
            if 'include' in tmpItem:
                includePatt = tmpItem['include'].split(',')
            if 'exclude' in tmpItem:
                excludePatt = tmpItem['exclude'].split(',')
    return fileList,includePatt,excludePatt



# append dataset
def appendDataset(taskParamMap,datasetSpec,fileList):
    # make item for dataset
    tmpItem = {}
    tmpItem['type']       = 'template'
    tmpItem['value']      = ''
    tmpItem['dataset']    = datasetSpec.datasetName
    tmpItem['param_type'] = datasetSpec.type
    if fileList != []:
        tmpItem['files']  = fileList
    # append
    if 'jobParameters' not in taskParamMap:
        taskParamMap['jobParameters'] = []
    taskParamMap['jobParameters'].append(tmpItem)
    return taskParamMap



# check if use random seed
def useRandomSeed(taskParamMap):
    for tmpItem in taskParamMap['jobParameters']:
        if 'value' in tmpItem:
            # get offset for random seed
            if tmpItem['type'] == 'template' and tmpItem['param_type'] == 'number':
                if '${RNDMSEED}' in tmpItem['value']:
                    return True
    return False



import sys
try:
    testTaskType = sys.argv[1]
except Exception:
    testTaskType = 'test'

try:
    vo = sys.argv[2]
except Exception:
    vo = 'atlas'

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jediddm.DDMInterface import DDMInterface

import multiprocessing

from pandajedi.jediorder import ContentsFeeder

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

ddmIF = DDMInterface()
ddmIF.setupInterface()

parent_conn, child_conn = multiprocessing.Pipe()

contentsFeeder = multiprocessing.Process(target=ContentsFeeder.launcher,
                                                args=(child_conn,tbIF,ddmIF,
                                                      vo,testTaskType))
contentsFeeder.start()



from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from .JobThrottlerBase import JobThrottlerBase

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# class to throttle ATLAS analysis jobs
class AtlasAnalJobThrottler (JobThrottlerBase):

    # constructor
    def __init__(self,taskBufferIF):
        JobThrottlerBase.__init__(self,taskBufferIF)


    # check if throttled
    def toBeThrottled(self, vo, prodSourceLabel, cloudName, workQueue, resourceType):
        # make logger
        tmpLog = MsgWrapper(logger)
        tmpLog.debug('start vo={0} label={1} cloud={2} workQueue={3}'.format(vo,prodSourceLabel,cloudName,
                                                                             workQueue.queue_name))
        # check if unthrottled
        if not workQueue.throttled:
            tmpLog.debug("  done : unthrottled since throttled is False")
            return self.retUnThrottled
        tmpLog.debug("  done : SKIP")
        return self.retThrottled



from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jediorder.JobThrottler import JobThrottler

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

vo = 'atlas'
prodSourceLabel = 'managed'
cloud = 'WORLD'

# get SiteMapper
siteMapper = tbIF.getSiteMapper()
wqMap = tbIF.getWorkQueueMap()

jt = JobThrottler(vo, prodSourceLabel)
jt.initializeMods(tbIF)

workQueues = wqMap.getAlignedQueueList(vo, prodSourceLabel)
resource_types = tbIF.load_resource_types()
for workQueue in workQueues:
    for resource_type in resource_types:
        print(jt.toBeThrottled(vo, prodSourceLabel, cloud, workQueue, resource_type.resource_name))



import uuid
try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from userinterface import Client

taskParamMap = {}

taskParamMap['noInput'] = True
taskParamMap['nEventsPerJob'] = 1
taskParamMap['nEvents'] = 5
taskParamMap['taskName'] = str(uuid.uuid4())
taskParamMap['userName'] = 'pandasrv1'
taskParamMap['vo'] = 'atlas'
taskParamMap['taskPriority'] = 1000
taskParamMap['reqID'] = 12345
taskParamMap['architecture'] = ''
taskParamMap['hpoWorkflow'] = True
taskParamMap['transUses'] = ''
taskParamMap['transHome'] = ''
taskParamMap['transPath'] = 'http://pandaserver.cern.ch:25080/trf/user/runHPO-00-00-01'
taskParamMap['processingType'] = 'simul'
taskParamMap['prodSourceLabel'] = 'test'
taskParamMap['useLocalIO'] = 1
taskParamMap['taskType'] = 'prod'
taskParamMap['workingGroup'] = 'AP_HPO'
taskParamMap['coreCount'] = 1
taskParamMap['site'] = 'BNL_PROD_UCORE'
taskParamMap['nucleus'] = 'CERN-PROD'
taskParamMap['cloud'] = 'WORLD'

logDatasetName = 'panda.jeditest.log.{0}'.format(uuid.uuid4())
outDatasetName = 'panda.jeditest.HPO.{0}'.format(uuid.uuid4())

taskParamMap['log'] = {'dataset': logDatasetName,
                       'type':'template',
                       'param_type':'log',
                       'token': 'ddd:.*DATADISK',
                       'destination':'(type=DATADISK)\(dontkeeplog=True)',
                       'offset':1000,
                       'value':'{0}.${{SN}}.log.tgz'.format(logDatasetName)}

taskParamMap['hpoRequestData'] = {'sandbox': 'gitlab-registry.cern.ch/zhangruihpc/steeringcontainer:latest',
                                  'executable': 'docker',
                                  'arguments': "/bin/bash -c \"hpogrid generate --n_point=%NUM_POINTS "
                                               "--max_point=%MAX_POINTS --infile=$PWD/%IN  --outfile=$PWD/%OUT "
                                               "-l=nevergrad\"",
                                  'output_json': 'output.json',
                                  'max_points': 10,
                                  'num_points_per_generation': 2,
                                  }

taskParamMap['container_name'] = 'docker://gitlab-registry.cern.ch/zhangruihpc/evaluationcontainer:mlflow'

taskParamMap['jobParameters'] = [
    {'type':'constant',
     'value': '-o out.json -j "" -p "{0}"'.format(quote('bash ./exec_in_container.sh'))
     },
    {'type': 'constant',
     'value': '--writeInputToTxt IN_DATA:input_ds.json --inSampleFile input_sample.json'
     },
    {'type': 'constant',
     'value': '-a aaa.tgz --sourceURL https://aipanda048.cern.ch:25443'
     },
    {'type': 'constant',
     'value': '--inMap "{\'IN_DATA\': ${IN_DATA/T}}"'
     },
    {'type':'template',
     'param_type':'input',
     'value':'-i "${IN_DATA/T}"',
     'dataset':'mc16_13TeV.501103.MGPy8EG_StauStauDirect_220p0_1p0_TFilt.merge.EVNT.e8102_e7400_tid21342682_00',
     'attributes': 'nosplit,repeat',
     },
    {'type': 'template',
     'param_type': 'output',
     'token': 'ATLASDATADISK',
     'value': '$JEDITASKID.metrics.${SN}.tgz',
     'dataset': outDatasetName,
     'hidden': True,
     },
    {'type': 'constant',
     'value': '--outMetricsFile=${OUTPUT0}^metrics.tgz',
     },
    ]

print(Client.insertTaskParams(taskParamMap))



import re
import os
import sys
import socket
import operator
import traceback

from six import iteritems

from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandajedi.jedirefine import RefinerUtils


# watchdog to take actions for jumbo jobs
class JumboWatchDog:

    # constructor
    def __init__(self, taskBufferIF, ddmIF, log, vo, prodSourceLabel):
        self.taskBufferIF = taskBufferIF
        self.ddmIF = ddmIF
        self.pid = '{0}-{1}_{2}-jumbo'.format(socket.getfqdn().split('.')[0], os.getpid(), os.getpgrp())
        self.log = log
        self.vo = vo
        self.prodSourceLabel = prodSourceLabel
        self.component = 'JumboWatchDog'
        self.dryRun = True


    # main
    def run(self):
        try:
            # get process lock
            locked = self.taskBufferIF.lockProcess_JEDI(vo=self.vo, prodSourceLabel=self.prodSourceLabel,
                                                        cloud=None, workqueue_id=None, resource_name=None,
                                                        component=self.component, pid=self.pid, timeLimit=10)
            if not locked:
                self.log.debug('component={0} skipped since locked by another'.format(self.component))
                return
            # get parameters for conversion
            self.log.debug('component={0} start'.format(self.component))
            maxTasks = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_MAX_TASKS', 'jedi', self.vo)
            if maxTasks is None:
                maxTasks = 1
            nEventsToDisable = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_MIN_EVENTS_DISABLE', 'jedi', self.vo)
            if nEventsToDisable is None:
                nEventsToDisable = 100000
            nEventsToEnable = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_MIN_EVENTS_ENABLE', 'jedi', self.vo)
            if nEventsToEnable is None:
                nEventsToEnable = nEventsToDisable * 10
            maxEvents = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_MAX_EVENTS', 'jedi', self.vo)
            if maxEvents is None:
                maxEvents = maxTasks * nEventsToEnable // 2
            nJumboPerTask = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_PER_TASK', 'jedi', self.vo)
            if nJumboPerTask is None:
                nJumboPerTask = 1
            nJumboPerSite = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_PER_SITE', 'jedi', self.vo)
            if nJumboPerSite is None:
                nJumboPerSite = 1
            maxPrio = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_MAX_CURR_PRIO', 'jedi', self.vo)
            if maxPrio is None:
                maxPrio = 500
            progressToBoost = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_PROG_TO_BOOST', 'jedi', self.vo)
            if progressToBoost is None:
                progressToBoost = 95
            maxFilesToBoost = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_MAX_FILES_TO_BOOST', 'jedi', self.vo)
            if maxFilesToBoost is None:
                maxFilesToBoost = 500
            prioToBoost = 900
            prioWhenDisabled = self.taskBufferIF.getConfigValue(self.component, 'JUMBO_PRIO_DISABLED', 'jedi', self.vo)
            if prioWhenDisabled is None:
                prioWhenDisabled = 500
            # get current info
            tasksWithJumbo = self.taskBufferIF.getTaskWithJumbo_JEDI(self.vo, self.prodSourceLabel)
            totEvents = 0
            doneEvents = 0
            nTasks = 0
            for jediTaskID, taskData in iteritems(tasksWithJumbo):
                # disable jumbo
                if taskData['useJumbo'] != JediTaskSpec.enum_useJumbo['disabled'] and taskData['site'] is None:
                    if  taskData['nEvents'] - taskData['nEventsDone'] < nEventsToDisable:
                        # disable
                        self.log.info('component={0} disable jumbo in jediTaskID={1} due to n_events_to_process={2} < {3}'.format(self.component, jediTaskID,
                                                                                                                                  taskData['nEvents'] - taskData['nEventsDone'],
                                                                                                                                  nEventsToDisable))
                        self.taskBufferIF.enableJumboJobs(jediTaskID, 0, 0)
                    else:
                        # wait
                        nTasks += 1
                        totEvents += taskData['nEvents']
                        doneEvents += taskData['nEventsDone']
                        self.log.info('component={0} keep jumbo in jediTaskID={1} due to n_events_to_process={2} > {3}'.format(self.component, jediTaskID,
                                                                                                                               taskData['nEvents'] - taskData['nEventsDone'],
                                                                                                                               nEventsToDisable))
                # increase priority for jumbo disabled
                if taskData['useJumbo'] == JediTaskSpec.enum_useJumbo['disabled'] and taskData['currentPriority'] < prioWhenDisabled:
                    self.taskBufferIF.changeTaskPriorityPanda(jediTaskID, prioWhenDisabled)
                    self.log.info('component={0} priority boost to {1} after disabing jumbo in in jediTaskID={2}'.format(self.component, prioWhenDisabled, jediTaskID))
                # increase priority when close to completion
                if taskData['nEvents'] > 0 and (taskData['nEvents'] - taskData['nEventsDone']) * 100 // taskData['nEvents'] < progressToBoost \
                        and taskData['currentPriority'] < prioToBoost and (taskData['nFiles'] - taskData['nFilesDone']) < maxFilesToBoost:
                    # boost
                    tmpStr = 'component={0} priority boost to {5} for jediTaskID={1} due to n_events_done={2} > {3}*{4}% '.format(self.component, jediTaskID,
                                                                                                                                  taskData['nEventsDone'],
                                                                                                                                  taskData['nEvents'],
                                                                                                                                  progressToBoost,
                                                                                                                                  prioToBoost)
                    tmpStr += 'n_files_remaining={0} < {1}'.format(taskData['nFiles'] - taskData['nFilesDone'], maxFilesToBoost)
                    self.log.info(tmpStr)
                    self.taskBufferIF.changeTaskPriorityPanda(jediTaskID, prioToBoost)
                # kick pending
                if taskData['taskStatus'] in ['pending', 'running'] and taskData['useJumbo'] in [JediTaskSpec.enum_useJumbo['pending'], JediTaskSpec.enum_useJumbo['running']]:
                    nActiveJumbo = 0
                    for computingSite, jobStatusMap in iteritems(taskData['jumboJobs']):
                        for jobStatus, nJobs in iteritems(jobStatusMap):
                            if jobStatus in ['defined', 'assigned', 'activated', 'sent', 'starting', 'running', 'transferring', 'holding']:
                                nActiveJumbo += nJobs
                    if nActiveJumbo == 0:
                        self.log.info('component={0} kick jumbo in {2} jediTaskID={1}'.format(self.component,
                                                                                              jediTaskID,
                                                                                              taskData['taskStatus']))
                        self.taskBufferIF.kickPendingTasksWithJumbo_JEDI(jediTaskID)
                # reset input to re-generate co-jumbo
                if taskData['currentPriority'] >= prioToBoost:
                    nReset = self.taskBufferIF.resetInputToReGenCoJumbo_JEDI(jediTaskID)
                    if nReset is not None and nReset > 0:
                        self.log.info('component={0} reset {1} inputs to regenerate co-jumbo for jediTaskID={2}'.format(self.component, nReset, jediTaskID))
                    else:
                        self.log.debug('component={0} tried to reset inputs to regenerate co-jumbo with {1} for jediTaskID={2}'.format(self.component, nReset, jediTaskID))
            self.log.info('component={0} total_events={1} n_events_to_process={2} n_tasks={3} available for jumbo'.format(self.component, totEvents,
                                                                                                                          totEvents - doneEvents, nTasks))
            if True:
                # get list of releases and caches available at jumbo job enabled PQs
                jumboRels, jumboCaches = self.taskBufferIF.getRelCacheForJumbo_JEDI()
                # look for tasks to enable jumbo
                if self.dryRun:
                    self.log.info('component={0} look for tasks to enable jumbo in dry run mode'.format(self.component))
                else:
                    self.log.info('component={0} look for tasks to enable jumbo due to lack of tasks and events to meet max_tasks={1} max_events={2}'.format(self.component,
                                                                                                                                                             maxTasks,
                                                                                                                                                             maxEvents))
                tasksToEnableJumbo = self.taskBufferIF.getTaskToEnableJumbo_JEDI(self.vo, self.prodSourceLabel, maxPrio, nEventsToEnable)
                nGoodTasks = 0
                self.log.debug('component={0} got {1} tasks to check'.format(self.component, len(tasksToEnableJumbo)))
                # sort by nevents
                nEventsMap = dict()
                for jediTaskID, taskData in iteritems(tasksToEnableJumbo):
                    nEventsMap[jediTaskID] = taskData['nEvents']
                sortedList = sorted(list(nEventsMap.items()), key=operator.itemgetter(1))
                sortedList.reverse()
                for jediTaskID, nEvents in sortedList:
                    taskData = tasksToEnableJumbo[jediTaskID]
                    # get task parameters
                    try:
                        taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(jediTaskID)
                        taskParamMap = RefinerUtils.decodeJSON(taskParam)
                    except Exception:
                        self.log.error('component={0} failed to get task params for jediTaskID={1}'.format(self.component, jediTaskID))
                        continue
                    taskSpec = JediTaskSpec()
                    taskSpec.splitRule = taskData['splitRule']
                    # check if good for jumbo
                    if 'esConvertible' not in taskParamMap or taskParamMap['esConvertible'] is False:
                        self.log.info('component={0} skip to enable jumbo for jediTaskID={1} since not ES-convertible'.format(self.component, jediTaskID))
                        continue
                    if taskSpec.inFilePosEvtNum():
                        pass
                    elif taskSpec.getNumFilesPerJob() == 1:
                        pass
                    elif taskSpec.getNumEventsPerJob() is not None and 'nEventsPerInputFile' in taskParamMap \
                            and taskSpec.getNumEventsPerJob() <= taskParamMap['nEventsPerInputFile']:
                        pass
                    else:
                        self.log.info('component={0} skip to enable jumbo for jediTaskID={1} since not good for in-file positional event numbers'.format(self.component, jediTaskID))
                        continue
                    # check software
                    transHome = taskData['transHome']
                    cmtConfig = taskData['architecture']
                    if re.search('^\d+\.\d+\.\d+$', transHome.split('-')[-1]) is not None:
                        transHome = transHome.split('-')[-1]
                        swDict = jumboRels
                    else:
                        swDict = jumboCaches
                    key = (transHome, cmtConfig)
                    if key not in swDict:
                        self.log.info('component={0} skip to enable jumbo for jediTaskID={1} since {2}:{3} is unavailable at jumbo job enabled PQs'.format(
                                self.component,
                                jediTaskID,
                                transHome,
                                cmtConfig))
                        continue
                    if not self.dryRun and nTasks < maxTasks and (totEvents - doneEvents) < maxEvents:
                        self.log.info('component={0} enable jumbo in jediTaskID={1} with n_events_to_process={2}'.format(self.component, jediTaskID,
                                                                                                                         taskData['nEvents'] - taskData['nEventsDone']))
                        if taskData['eventService'] == 0:
                            tmpS, tmpO = self.taskBufferIF.enableEventService(taskData['jediTaskID'])
                            if tmpS != 0:
                                self.log.error('component={0} failed to enable ES in jediTaskID={1} with {2}'.format(self.component, jediTaskID, tmpO))
                                continue
                        self.taskBufferIF.enableJumboJobs(taskData['jediTaskID'], nJumboPerTask, nJumboPerSite)
                        nTasks += 1
                        totEvents += taskData['nEvents']
                        doneEvents += taskData['nEventsDone']
                    else:
                        nGoodTasks += 1
                        self.log.info('component={0} good to enable jumbo in jediTaskID={1} with n_events_to_process={2}'.format(self.component, jediTaskID,
                                                                                                                                 taskData['nEvents'] - taskData['nEventsDone']))
                self.log.info('component={0} there are n_good_tasks={1} tasks good for jumbo'.format(self.component, nGoodTasks))
            self.log.debug('component={0} done'.format(self.component))
        except Exception:
            # error
            errtype, errvalue = sys.exc_info()[:2]
            errStr = ": %s %s" % (errtype.__name__, errvalue)
            errStr.strip()
            errStr += traceback.format_exc()
            self.log.error(errStr)



import os
import sys
import re
import socket
import traceback

from six import iteritems

from .WatchDogBase import WatchDogBase
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.ThreadUtils import ListWithLock
from pandajedi.jedicore import Interaction
from pandajedi.jedibrokerage import AtlasBrokerUtils

from pandaserver.dataservice import DataServiceUtils
# from pandaserver.dataservice.Activator import Activator
from pandaserver.taskbuffer import JobUtils

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# task withholder watchdog for ATLAS
class AtlasTaskWithholderWatchDog(WatchDogBase):

    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        WatchDogBase.__init__(self, taskBufferIF, ddmIF)
        self.pid = '{0}-{1}-dog'.format(socket.getfqdn().split('.')[0], os.getpid())
        # self.cronActions = {'forPrestage': 'atlas_prs'}
        self.vo = 'atlas'
        self.prodSourceLabelList = ['managed']
        # call refresh
        self.refresh()

    # refresh information stored in the instance
    def refresh(self):
        # work queue mapper
        self.workQueueMapper = self.taskBufferIF.getWorkQueueMap()
        # site mapper
        self.siteMapper = self.taskBufferIF.getSiteMapper()
        # all sites
        allSiteList = []
        for siteName, tmpSiteSpec in iteritems(self.siteMapper.siteSpecList):
            # if tmpSiteSpec.type == 'analysis' or tmpSiteSpec.is_grandly_unified():
            allSiteList.append(siteName)
        self.allSiteList = allSiteList

    # get map of site to list of RSEs and blacklisted RSEs
    def get_site_rse_map_and_blacklisted_rse_set(self, prod_source_label):
        site_rse_map = {}
        blacklisted_rse_set = set()
        for tmpPseudoSiteName in self.allSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpPseudoSiteName)
            tmpSiteName = tmpSiteSpec.get_unified_name()
            scope_input, scope_output = DataServiceUtils.select_scope(tmpSiteSpec, prod_source_label, prod_source_label)
            try:
                tmp_ddm_spec = tmpSiteSpec.ddm_endpoints_input[scope_input]
                endpoint_name = tmpSiteSpec.ddm_input[scope_input]
                endpoint_token_map = tmp_ddm_spec.getTokenMap('input')
                tmp_endpoint = tmp_ddm_spec.getEndPoint(endpoint_name)
            except KeyError:
                continue
            else:
                # fill site rse map
                site_rse_map[tmpSiteName] = list(endpoint_token_map.values())
                # blacklisted rse
                if tmp_endpoint is not None and tmp_endpoint['blacklisted'] == 'Y':
                    blacklisted_rse_set.add(endpoint_name)
        # return
        return site_rse_map, blacklisted_rse_set

    # get busy sites
    def get_busy_sites(self, gshare, cutoff):
        busy_sites_list = []
        # get global share
        tmpSt, jobStatPrioMap = self.taskBufferIF.getJobStatisticsByGlobalShare(self.vo)
        if not tmpSt:
            # got nothing...
            return busy_sites_list
        for tmpPseudoSiteName in self.allSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpPseudoSiteName)
            tmpSiteName = tmpSiteSpec.get_unified_name()
            # get nQueue and nRunning
            nRunning = AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, 'running', gshare)
            nQueue = 0
            for jobStatus in ['defined', 'assigned', 'activated', 'starting']:
                nQueue += AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, jobStatus, gshare)
            # busy sites
            if nQueue > max(cutoff, nRunning*2):
                busy_sites_list.append(tmpSiteName)
        # return
        return busy_sites_list

    # # handle waiting jobs
    # def do_make_tasks_pending(self, task_list):
    #     tmpLog = MsgWrapper(logger, 'do_make_tasks_pending')
    #     tmpLog.debug('start')
    #     # check every x min
    #     checkInterval = 20
    #     # make task pending
    #     for taskID, pending_reason in task_list:
    #         tmpLog = MsgWrapper(logger, '< #ATM #KV do_make_tasks_pending jediTaskID={0}>'.format(taskID))
    #         retVal = self.taskBufferIF.makeTaskPending_JEDI(taskID, reason=pending_reason)
    #         tmpLog.debug('done with {0}'.format(retVal))


    # set tasks to be pending due to condition of data locality
    def do_for_data_locality(self):
        tmp_log = MsgWrapper(logger)
        # refresh
        self.refresh()
        # list of resource type
        # resource_type_list = [ rt.resource_name for rt in self.taskBufferIF.load_resource_types() ]
        # loop
        for prod_source_label in self.prodSourceLabelList:
            # site-rse map and blacklisted rses
            site_rse_map, blacklisted_rse_set = self.get_site_rse_map_and_blacklisted_rse_set(prod_source_label)
            tmp_log.debug('Found {0} blacklisted RSEs : {1}'.format(len(blacklisted_rse_set), ','.join(list(blacklisted_rse_set))))
            # parameter from GDP config
            upplimit_ioIntensity = self.taskBufferIF.getConfigValue(
                                        'task_withholder', 'LIMIT_IOINTENSITY_{0}'.format(prod_source_label), 'jedi', self.vo)
            lowlimit_currentPriority = self.taskBufferIF.getConfigValue(
                                        'task_withholder', 'LIMIT_PRIORITY_{0}'.format(prod_source_label), 'jedi', self.vo)
            if upplimit_ioIntensity is None:
                upplimit_ioIntensity = 999999
            if lowlimit_currentPriority is None:
                lowlimit_currentPriority = -999999
            upplimit_ioIntensity = max(upplimit_ioIntensity, 100)
            # get work queue for gshare
            work_queue_list = self.workQueueMapper.getAlignedQueueList(self.vo, prod_source_label)
            # loop over work queue
            for work_queue in work_queue_list:
                gshare = work_queue.queue_name
                # get cutoff
                cutoff = self.taskBufferIF.getConfigValue('jobbroker', 'NQUEUELIMITSITE_{}'.format(gshare),
                                                          'jedi', self.vo)
                if not cutoff:
                    cutoff = 20
                # busy sites
                busy_sites_list = self.get_busy_sites(gshare, cutoff)
                # rses of busy sites
                busy_rses = set()
                for site in busy_sites_list:
                    try:
                        busy_rses.update(set(site_rse_map[site]))
                    except KeyError:
                        continue
                # make sql parameters of rses
                to_exclude_rses = list(busy_rses|blacklisted_rse_set)
                rse_params_list = []
                rse_params_map = {}
                for j, rse in enumerate(to_exclude_rses):
                    rse_param = ':rse_{0}'.format(j + 1)
                    rse_params_list.append(rse_param)
                    rse_params_map[rse_param] = rse
                rse_params_str = ','.join(rse_params_list)
                # sql
                sql_query = (
                    "SELECT t.jediTaskID "
                    "FROM {jedi_schema}.JEDI_Tasks t "
                    "WHERE t.status IN ('ready','running','scouting') AND t.lockedBy IS NULL "
                        "AND t.gshare=:gshare "
                        "AND t.ioIntensity>=:ioIntensity AND t.currentPriority<:currentPriority "
                        "AND EXISTS ( "
                            "SELECT * FROM {jedi_schema}.JEDI_Datasets d "
                            "WHERE d.jediTaskID=t.jediTaskID "
                                "AND d.type='input' "
                            ") "
                        "AND NOT EXISTS ( "
                            "SELECT * FROM {jedi_schema}.JEDI_Dataset_Locality dl "
                            "WHERE dl.jediTaskID=t.jediTaskID "
                                "AND dl.rse NOT IN ({rse_params_str}) "
                            ") "
                    "FOR UPDATE "
                ).format(jedi_schema=jedi_config.db.schemaJEDI, rse_params_str=rse_params_str)
                # params map
                params_map = {
                        ':gshare': gshare,
                        ':ioIntensity': upplimit_ioIntensity,
                        ':currentPriority': lowlimit_currentPriority,
                    }
                params_map.update(rse_params_map)
                # pending reason
                reason = 'no local input data, ioIntensity>={ioIntensity}, currentPriority<{currentPriority},'\
                         'nQueue>max({cutOff},nRunning*2) at all sites where the task can run'.format(
                    ioIntensity=upplimit_ioIntensity,currentPriority=lowlimit_currentPriority,
                    cutOff=cutoff)
                # set pending
                dry_run = False
                if dry_run:
                    dry_sql_query = (
                        "SELECT t.jediTaskID "
                        "FROM {jedi_schema}.JEDI_Tasks t "
                        "WHERE t.status IN ('ready','running','scouting') AND t.lockedBy IS NULL "
                            "AND t.gshare=:gshare "
                            "AND t.ioIntensity>=:ioIntensity AND t.currentPriority<:currentPriority "
                            "AND EXISTS ( "
                                "SELECT * FROM {jedi_schema}.JEDI_Datasets d "
                                "WHERE d.jediTaskID=t.jediTaskID "
                                    "AND d.type='input' "
                                ") "
                            "AND NOT EXISTS ( "
                                "SELECT * FROM {jedi_schema}.JEDI_Dataset_Locality dl "
                                "WHERE dl.jediTaskID=t.jediTaskID "
                                    "AND dl.rse NOT IN ({rse_params_str}) "
                                ") "
                    ).format(jedi_schema=jedi_config.db.schemaJEDI, rse_params_str=rse_params_str)
                    res = self.taskBufferIF.querySQL(dry_sql_query, params_map)
                    n_tasks = 0 if res is None else len(res)
                    if n_tasks > 0:
                        result = [ x[0] for x in res ]
                        tmp_log.debug('[dry run] gshare: {gshare:<16} {n_tasks:>5} tasks would be pending : {result} ; reason="{reason}" '.format(
                                        gshare=gshare, n_tasks=n_tasks, result=result, reason=reason))
                else:
                    n_tasks = self.taskBufferIF.queryTasksToBePending_JEDI(sql_query, params_map, reason)
                    if n_tasks is not None and n_tasks > 0:
                        tmp_log.info('gshare: {gshare:<16} {n_tasks:>5} tasks got pending ; reason="{reason}" '.format(
                                        gshare=gshare, n_tasks=str(n_tasks), reason=reason))


    # main
    def doAction(self):
        try:
            # get logger
            origTmpLog = MsgWrapper(logger)
            origTmpLog.debug('start')
            # make tasks pending under certain conditions
            self.do_for_data_locality()
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            err_str = traceback.format_exc()
            origTmpLog.error('failed with {0} {1} ; {2}'.format(errtype, errvalue, err_str))
        # return
        origTmpLog.debug('done')
        return self.SC_SUCCEEDED



import os
import re
import sys
import time
import uuid
import math
import socket
import datetime
import traceback

from six import iteritems

from pandajedi.jedicore.ThreadUtils import ListWithLock,ThreadPool,WorkerThread
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedirefine import RefinerUtils
from .JediKnight import JediKnight

from pandajedi.jediconfig import jedi_config

try:
    from idds.client.client import Client as iDDS_Client
    import idds.common.constants
    import idds.common.utils
except ImportError:
    pass


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# worker class to take care of DatasetContents table
class ContentsFeeder (JediKnight):
    # constructor
    def __init__(self,commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels):
        self.vos = self.parseInit(vos)
        self.prodSourceLabels = self.parseInit(prodSourceLabels)
        self.pid = '{0}-{1}_{2}-con'.format(socket.getfqdn().split('.')[0],os.getpid(),os.getpgrp())
        JediKnight.__init__(self,commuChannel,taskBufferIF,ddmIF,logger)


    # main
    def start(self):
        # start base class
        JediKnight.start(self)
        # go into main loop
        while True:
            startTime = datetime.datetime.utcnow()
            try:
                # loop over all vos
                for vo in self.vos:
                    # loop over all sourceLabels
                    for prodSourceLabel in self.prodSourceLabels:
                        # get the list of datasets to feed contents to DB
                        tmpList = self.taskBufferIF.getDatasetsToFeedContents_JEDI(vo,prodSourceLabel)
                        if tmpList is None:
                            # failed
                            logger.error('failed to get the list of datasets to feed contents')
                        else:
                            logger.debug('got %s datasets' % len(tmpList))
                            # put to a locked list
                            dsList = ListWithLock(tmpList)
                            # make thread pool
                            threadPool = ThreadPool()
                            # make workers
                            nWorker = jedi_config.confeeder.nWorkers
                            for iWorker in range(nWorker):
                                thr = ContentsFeederThread(dsList,threadPool,
                                                           self.taskBufferIF,self.ddmIF,
                                                           self.pid)
                                thr.start()
                            # join
                            threadPool.join()
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                logger.error('failed in %s.start() with %s %s' % (self.__class__.__name__,errtype.__name__,errvalue))
            # sleep if needed
            loopCycle = jedi_config.confeeder.loopCycle
            timeDelta = datetime.datetime.utcnow() - startTime
            sleepPeriod = loopCycle - timeDelta.seconds
            if sleepPeriod > 0:
                time.sleep(sleepPeriod)
            # randomize cycle
            self.randomSleep(max_val=loopCycle)



# thread for real worker
class ContentsFeederThread (WorkerThread):

    # constructor
    def __init__(self,taskDsList,threadPool,taskbufferIF,ddmIF,pid):
        # initialize woker with no semaphore
        WorkerThread.__init__(self,None,threadPool,logger)
        # attributres
        self.taskDsList = taskDsList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF
        self.msgType = 'contentsfeeder'
        self.pid     = pid



    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 10
                taskDsList = self.taskDsList.get(nTasks)
                # no more datasets
                if len(taskDsList) == 0:
                    self.logger.debug('%s terminating since no more items' % self.__class__.__name__)
                    return
                # loop over all tasks
                for jediTaskID,dsList in taskDsList:
                    allUpdated = True
                    taskBroken = False
                    taskOnHold = False
                    runningTask = False
                    missingMap = {}
                    datasetsIdxConsistency = []

                    # get task
                    tmpStat,taskSpec = self.taskBufferIF.getTaskWithID_JEDI(jediTaskID, False, True, self.pid, 10,
                                                                            clearError=True)
                    if not tmpStat or taskSpec is None:
                        self.logger.debug('failed to get taskSpec for jediTaskID={0}'.format(jediTaskID))
                        continue

                    # make logger
                    try:
                        gshare = '_'.join(taskSpec.gshare.split(' '))
                    except Exception:
                        gshare = 'Undefined'
                    tmpLog = MsgWrapper(self.logger,'<jediTaskID={0} gshare={1}>'.format(jediTaskID, gshare))

                    try:
                        # get task parameters
                        taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(jediTaskID)
                        taskParamMap = RefinerUtils.decodeJSON(taskParam)
                    except Exception as e:
                        tmpLog.error('task param conversion from json failed with {}'.format(str(e)))
                        # unlock
                        tmpStat = self.taskBufferIF.unlockSingleTask_JEDI(jediTaskID, self.pid)
                        tmpLog.debug('unlocked with {}'.format(tmpStat))
                        continue
                    # renaming of parameters
                    if 'nEventsPerInputFile' in taskParamMap:
                        taskParamMap['nEventsPerFile'] = taskParamMap['nEventsPerInputFile']
                    # the number of files per job
                    nFilesPerJob = taskSpec.getNumFilesPerJob()
                    # the number of chunks used by scout
                    nChunksForScout = 10
                    # load XML
                    if taskSpec.useLoadXML():
                        xmlConfig = taskParamMap['loadXML']
                    else:
                        xmlConfig = None
                    # skip files used by another task
                    if 'skipFilesUsedBy' in taskParamMap:
                        skipFilesUsedBy = taskParamMap['skipFilesUsedBy']
                    else:
                        skipFilesUsedBy = None
                    # check no wait
                    noWaitParent = False
                    parentOutDatasets = set()
                    if taskSpec.noWaitParent() and taskSpec.parent_tid not in [None,taskSpec.jediTaskID]:
                        tmpStat = self.taskBufferIF.checkParentTask_JEDI(taskSpec.parent_tid)
                        if tmpStat is None or tmpStat == 'running':
                            noWaitParent = True
                            # get output datasets from parent task
                            tmpParentStat,tmpParentOutDatasets = self.taskBufferIF.getDatasetsWithJediTaskID_JEDI(taskSpec.parent_tid,
                                                                                                                  ['output','log'])
                            # collect dataset names
                            for tmpParentOutDataset in tmpParentOutDatasets:
                                parentOutDatasets.add(tmpParentOutDataset.datasetName)
                                if tmpParentOutDataset.containerName:
                                    if tmpParentOutDataset.containerName.endswith('/'):
                                        parentOutDatasets.add(tmpParentOutDataset.containerName)
                                        parentOutDatasets.add(tmpParentOutDataset.containerName[:-1])
                                    else:
                                        parentOutDatasets.add(tmpParentOutDataset.containerName)
                                        parentOutDatasets.add(tmpParentOutDataset.containerName+'/')
                    # loop over all datasets
                    nFilesMaster = 0
                    checkedMaster = False
                    setFrozenTime = True
                    if not taskBroken:
                        ddmIF = self.ddmIF.getInterface(taskSpec.vo, taskSpec.cloud)
                        origNumFiles = None
                        if 'nFiles' in taskParamMap:
                            origNumFiles = taskParamMap['nFiles']
                        id_to_container = {}
                        [id_to_container.update({datasetSpec.datasetID: datasetSpec.containerName}) \
                            for datasetSpec in dsList]
                        for datasetSpec in dsList:
                            tmpLog.debug('start loop for {0}(id={1})'.format(datasetSpec.datasetName,datasetSpec.datasetID))
                            # index consistency
                            if datasetSpec.indexConsistent():
                                datasetsIdxConsistency.append(datasetSpec.datasetID)
                            # prestaging
                            if taskSpec.inputPreStaging() and (datasetSpec.isMaster() or datasetSpec.isSeqNumber()):
                                nStaging = self.taskBufferIF.getNumStagingFiles_JEDI(taskSpec.jediTaskID)
                                if nStaging is not None and nStaging == 0 and datasetSpec.nFiles > 0:
                                    inputPreStaging = False
                                else:
                                    inputPreStaging = True
                            else:
                                inputPreStaging = False
                            # get dataset metadata
                            tmpLog.debug('get metadata')
                            gotMetadata = False
                            stateUpdateTime = datetime.datetime.utcnow()
                            try:
                                if not datasetSpec.isPseudo():
                                    tmpMetadata = ddmIF.getDatasetMetaData(datasetSpec.datasetName, ignore_missing=True)
                                else:
                                    # dummy metadata for pseudo dataset
                                    tmpMetadata = {'state':'closed'}
                                # set mutable when the dataset is open and parent is running or task is configured to run until the dataset is closed
                                if (noWaitParent or taskSpec.runUntilClosed() or inputPreStaging) and \
                                        (tmpMetadata['state'] == 'open'
                                        or datasetSpec.datasetName in parentOutDatasets
                                        or datasetSpec.datasetName.split(':')[-1] in parentOutDatasets
                                        or inputPreStaging):
                                    # dummy metadata when parent is running
                                    tmpMetadata = {'state':'mutable'}
                                gotMetadata = True
                            except Exception:
                                errtype,errvalue = sys.exc_info()[:2]
                                tmpLog.error('{0} failed to get metadata to {1}:{2}'.format(self.__class__.__name__,
                                                                                            errtype.__name__,errvalue))
                                if errtype == Interaction.JEDIFatalError:
                                    # fatal error
                                    datasetStatus = 'broken'
                                    taskBroken = True
                                    # update dataset status
                                    self.updateDatasetStatus(datasetSpec,datasetStatus,tmpLog)
                                else:
                                    if not taskSpec.ignoreMissingInDS():
                                        # temporary error
                                        taskOnHold = True
                                    else:
                                        # ignore missing
                                        datasetStatus = 'failed'
                                        # update dataset status
                                        self.updateDatasetStatus(datasetSpec,datasetStatus,tmpLog)
                                taskSpec.setErrDiag('failed to get metadata for {0}'.format(datasetSpec.datasetName))
                                if not taskSpec.ignoreMissingInDS():
                                    allUpdated = False
                            else:
                                # to skip missing dataset
                                if tmpMetadata['state'] == 'missing':
                                    # ignore missing
                                    datasetStatus = 'finished'
                                    # update dataset status
                                    self.updateDatasetStatus(datasetSpec, datasetStatus, tmpLog, 'closed')
                                    tmpLog.debug('disabled missing {0}'.format(datasetSpec.datasetName))
                                    continue
                                # get file list specified in task parameters
                                if taskSpec.is_work_segmented() and not datasetSpec.isPseudo() and \
                                        not datasetSpec.isMaster():
                                    fileList = []
                                    includePatt = []
                                    excludePatt = []
                                    try:
                                        segment_id = int(id_to_container[datasetSpec.masterID].split('/')[-1])
                                        for item in taskParamMap['segmentSpecs']:
                                            if item['id'] == segment_id:
                                                if 'files' in item:
                                                    fileList = item['files']
                                                elif 'datasets' in item:
                                                    for tmpDatasetName in item['datasets']:
                                                        tmpRet = ddmIF.getFilesInDataset(tmpDatasetName)
                                                        fileList += [tmpAttr['lfn'] for tmpAttr in tmpRet.values()]
                                    except Exception:
                                        pass
                                else:
                                    fileList,includePatt,excludePatt = RefinerUtils.extractFileList(
                                        taskParamMap, datasetSpec.datasetName)
                                # get the number of events in metadata
                                if 'getNumEventsInMetadata' in taskParamMap:
                                    getNumEvents = True
                                else:
                                    getNumEvents = False
                                # get file list from DDM
                                tmpLog.debug('get files')
                                try:
                                    useInFilesWithNewAttemptNr = False
                                    skipDuplicate = not datasetSpec.useDuplicatedFiles()
                                    if not datasetSpec.isPseudo():
                                        if fileList != [] and 'useInFilesInContainer' in taskParamMap and \
                                                datasetSpec.containerName not in ['',None]:
                                            # read files from container if file list is specified in task parameters
                                            tmpDatasetName = datasetSpec.containerName
                                        else:
                                            tmpDatasetName = datasetSpec.datasetName
                                        # use long format for LB
                                        longFormat = False
                                        if taskSpec.respectLumiblock() or taskSpec.orderByLB():
                                            longFormat = True
                                        tmpRet = ddmIF.getFilesInDataset(tmpDatasetName,
                                                                         getNumEvents=getNumEvents,
                                                                         skipDuplicate=skipDuplicate,
                                                                         longFormat=longFormat
                                                                         )
                                        tmpLog.debug('got {0} files in {1}'.format(len(tmpRet),tmpDatasetName))
                                        # remove lost files
                                        """
                                        tmpLostFiles = ddmIF.findLostFiles(tmpDatasetName,tmpRet)
                                        if tmpLostFiles != {}:
                                            tmpLog.debug('found {0} lost files in {1}'.format(len(tmpLostFiles),tmpDatasetName))
                                            for tmpListGUID,tmpLostLFN in iteritems(tmpLostFiles):
                                                tmpLog.debug('removed {0}'.format(tmpLostLFN))
                                                del tmpRet[tmpListGUID]
                                        """
                                    else:
                                        if datasetSpec.isSeqNumber():
                                            # make dummy files for seq_number
                                            if datasetSpec.getNumRecords() is not None:
                                                nPFN = datasetSpec.getNumRecords()
                                            elif origNumFiles is not None:
                                                nPFN = origNumFiles
                                                if 'nEventsPerJob' in taskParamMap and 'nEventsPerFile' in taskParamMap \
                                                        and taskParamMap['nEventsPerFile'] > taskParamMap['nEventsPerJob']:
                                                    nPFN = nPFN * math.ceil(taskParamMap['nEventsPerFile'] / taskParamMap['nEventsPerJob'])
                                                elif 'nEventsPerFile' in taskParamMap and 'nEventsPerRange' in taskParamMap:
                                                    nPFN = nPFN * taskParamMap['nEventsPerFile'] // taskParamMap['nEventsPerRange']
                                            elif 'nEvents' in taskParamMap and 'nEventsPerJob' in taskParamMap:
                                                nPFN = math.ceil(taskParamMap['nEvents'] / taskParamMap['nEventsPerJob'])
                                            elif 'nEvents' in taskParamMap and 'nEventsPerFile' in taskParamMap \
                                                    and taskSpec.getNumFilesPerJob() is not None:
                                                nPFN = math.ceil(taskParamMap['nEvents'] / taskParamMap['nEventsPerFile'] / taskSpec.getNumFilesPerJob())
                                            else:
                                                # the default number of records for seq_number
                                                seqDefNumRecords = 10000
                                                # get nFiles of the master
                                                tmpMasterAtt = self.taskBufferIF.getDatasetAttributes_JEDI(datasetSpec.jediTaskID,
                                                                                                           datasetSpec.masterID,
                                                                                                           ['nFiles'])
                                                # use nFiles of the master as the number of records if it is larger than the default
                                                if 'nFiles' in tmpMasterAtt and tmpMasterAtt['nFiles'] > seqDefNumRecords:
                                                    nPFN = tmpMasterAtt['nFiles']
                                                else:
                                                    nPFN = seqDefNumRecords
                                                # check usedBy
                                                if skipFilesUsedBy is not None:
                                                    for tmpJediTaskID in str(skipFilesUsedBy).split(','):
                                                        tmpParentAtt = self.taskBufferIF.getDatasetAttributesWithMap_JEDI(tmpJediTaskID,
                                                                                                                          {'datasetName':datasetSpec.datasetName},
                                                                                                                          ['nFiles'])
                                                        if 'nFiles' in tmpParentAtt and tmpParentAtt['nFiles']:
                                                            nPFN += tmpParentAtt['nFiles']
                                            tmpRet = {}
                                            # get offset
                                            tmpOffset = datasetSpec.getOffset()
                                            tmpOffset += 1
                                            for iPFN in range(nPFN):
                                                tmpRet[str(uuid.uuid4())] = {'lfn':iPFN+tmpOffset,
                                                                             'scope':None,
                                                                             'filesize':0,
                                                                             'checksum':None,
                                                                             }
                                        elif not taskSpec.useListPFN():
                                            # dummy file list for pseudo dataset
                                            tmpRet = {str(uuid.uuid4()):{'lfn':'pseudo_lfn',
                                                                         'scope':None,
                                                                         'filesize':0,
                                                                         'checksum':None,
                                                                         }
                                                      }
                                        else:
                                            # make dummy file list for PFN list
                                            if 'nFiles' in taskParamMap:
                                                nPFN = taskParamMap['nFiles']
                                            else:
                                                nPFN = 1
                                            tmpRet = {}
                                            for iPFN in range(nPFN):
                                                tmpRet[str(uuid.uuid4())] = {'lfn':'{0:06d}:{1}'.format(iPFN,taskParamMap['pfnList'][iPFN].split('/')[-1]),
                                                                             'scope':None,
                                                                             'filesize':0,
                                                                             'checksum':None,
                                                                             }
                                except Exception:
                                    errtype,errvalue = sys.exc_info()[:2]
                                    tmpLog.error('failed to get files due to {0}:{1} {2}'.format(self.__class__.__name__,
                                                                                                 errtype.__name__,errvalue))
                                    if errtype == Interaction.JEDIFatalError:
                                        # fatal error
                                        datasetStatus = 'broken'
                                        taskBroken = True
                                        # update dataset status
                                        self.updateDatasetStatus(datasetSpec,datasetStatus,tmpLog)
                                    else:
                                        # temporary error
                                        taskOnHold = True
                                    taskSpec.setErrDiag('failed to get files for {0}'.format(datasetSpec.datasetName))
                                    allUpdated = False
                                else:
                                    # parameters for master input
                                    respectLB = False
                                    useRealNumEvents = False
                                    if datasetSpec.isMaster():
                                        # respect LB boundaries
                                        respectLB = taskSpec.respectLumiblock()
                                        # use real number of events
                                        useRealNumEvents = taskSpec.useRealNumEvents()
                                    # the number of events per file
                                    nEventsPerFile  = None
                                    nEventsPerJob   = None
                                    nEventsPerRange = None
                                    tgtNumEventsPerJob = None
                                    if (datasetSpec.isMaster() and ('nEventsPerFile' in taskParamMap or useRealNumEvents)) or \
                                            (datasetSpec.isPseudo() and 'nEvents' in taskParamMap and not datasetSpec.isSeqNumber()):
                                        if 'nEventsPerFile' in taskParamMap:
                                            nEventsPerFile = taskParamMap['nEventsPerFile']
                                        elif datasetSpec.isMaster() and datasetSpec.isPseudo() and 'nEvents' in taskParamMap:
                                            # use nEvents as nEventsPerFile for pseudo input
                                            nEventsPerFile = taskParamMap['nEvents']
                                        if 'nEventsPerJob' in taskParamMap:
                                            nEventsPerJob = taskParamMap['nEventsPerJob']
                                        elif 'nEventsPerRange' in taskParamMap:
                                            nEventsPerRange = taskParamMap['nEventsPerRange']
                                        if 'tgtNumEventsPerJob' in taskParamMap:
                                            tgtNumEventsPerJob = taskParamMap['tgtNumEventsPerJob']
                                            # reset nEventsPerJob
                                            nEventsPerJob = None
                                    # max attempts
                                    maxAttempt = None
                                    maxFailure = None
                                    if datasetSpec.isMaster() or datasetSpec.toKeepTrack():
                                        # max attempts
                                        if taskSpec.disableAutoRetry():
                                            # disable auto retry
                                            maxAttempt = 1
                                        elif 'maxAttempt' in taskParamMap:
                                            maxAttempt = taskParamMap['maxAttempt']
                                        else:
                                            # use default value
                                            maxAttempt = 3
                                        # max failure
                                        if 'maxFailure' in taskParamMap:
                                            maxFailure = taskParamMap['maxFailure']
                                    # first event number
                                    firstEventNumber = None
                                    if datasetSpec.isMaster():
                                        # first event number
                                        firstEventNumber = 1 + taskSpec.getFirstEventOffset()
                                    # nMaxEvents
                                    nMaxEvents = None
                                    if datasetSpec.isMaster() and 'nEvents' in taskParamMap:
                                        nMaxEvents = taskParamMap['nEvents']
                                    # nMaxFiles
                                    nMaxFiles = None
                                    if 'nFiles' in taskParamMap:
                                        if datasetSpec.isMaster():
                                            nMaxFiles = taskParamMap['nFiles']
                                        else:
                                            # calculate for secondary
                                            if not datasetSpec.isPseudo():
                                                # check nFilesPerJob
                                                nFilesPerJobSec = datasetSpec.getNumFilesPerJob()
                                                if nFilesPerJobSec is not None:
                                                    nMaxFiles = origNumFiles * nFilesPerJobSec
                                            # check ratio
                                            if nMaxFiles is None:
                                                nMaxFiles = datasetSpec.getNumMultByRatio(origNumFiles)
                                            # multiplied by the number of jobs per file for event-level splitting
                                            if nMaxFiles is not None:
                                                if 'nEventsPerFile' in taskParamMap:
                                                    if 'nEventsPerJob' in taskParamMap:
                                                        if taskParamMap['nEventsPerFile'] > taskParamMap['nEventsPerJob']:
                                                            nMaxFiles *= float(taskParamMap['nEventsPerFile'])/float(taskParamMap['nEventsPerJob'])
                                                            nMaxFiles = int(math.ceil(nMaxFiles))
                                                    elif 'nEventsPerRange' in taskParamMap:
                                                        if taskParamMap['nEventsPerFile'] > taskParamMap['nEventsPerRange']:
                                                            nMaxFiles *= float(taskParamMap['nEventsPerFile'])/float(taskParamMap['nEventsPerRange'])
                                                            nMaxFiles = int(math.ceil(nMaxFiles))
                                                elif 'useRealNumEvents' in taskParamMap:
                                                    # reset nMaxFiles since it is unknown
                                                    nMaxFiles = None
                                    # use scout
                                    useScout = False
                                    if datasetSpec.isMaster() and taskSpec.useScout() and (datasetSpec.status != 'toupdate' or not taskSpec.isPostScout()):
                                        useScout = True
                                    # use files with new attempt numbers
                                    useFilesWithNewAttemptNr = False
                                    if not datasetSpec.isPseudo() and fileList != [] and 'useInFilesWithNewAttemptNr' in taskParamMap:
                                        useFilesWithNewAttemptNr = True
                                    # ramCount
                                    ramCount = 0
                                    # skip short input
                                    if datasetSpec.isMaster() and not datasetSpec.isPseudo() \
                                            and nEventsPerFile is not None and nEventsPerJob is not None \
                                            and nEventsPerFile >= nEventsPerJob \
                                            and 'skipShortInput' in taskParamMap and taskParamMap['skipShortInput'] is True:
                                        skipShortInput = True
                                    else:
                                        skipShortInput = False
                                    # feed files to the contents table
                                    tmpLog.debug('update contents')
                                    retDB,missingFileList,nFilesUnique,diagMap = \
                                        self.taskBufferIF.insertFilesForDataset_JEDI(datasetSpec,tmpRet,
                                                                                     tmpMetadata['state'],
                                                                                     stateUpdateTime,
                                                                                     nEventsPerFile,
                                                                                     nEventsPerJob,
                                                                                     maxAttempt,
                                                                                     firstEventNumber,
                                                                                     nMaxFiles,
                                                                                     nMaxEvents,
                                                                                     useScout,
                                                                                     fileList,
                                                                                     useFilesWithNewAttemptNr,
                                                                                     nFilesPerJob,
                                                                                     nEventsPerRange,
                                                                                     nChunksForScout,
                                                                                     includePatt,
                                                                                     excludePatt,
                                                                                     xmlConfig,
                                                                                     noWaitParent,
                                                                                     taskSpec.parent_tid,
                                                                                     self.pid,
                                                                                     maxFailure,
                                                                                     useRealNumEvents,
                                                                                     respectLB,
                                                                                     tgtNumEventsPerJob,
                                                                                     skipFilesUsedBy,
                                                                                     ramCount,
                                                                                     taskSpec,
                                                                                     skipShortInput,
                                                                                     inputPreStaging)
                                    if retDB is False:
                                        taskSpec.setErrDiag('failed to insert files for {0}. {1}'.format(datasetSpec.datasetName,
                                                                                                         diagMap['errMsg']))
                                        allUpdated = False
                                        taskBroken = True
                                        break
                                    elif retDB is None:
                                        # the dataset is locked by another or status is not applicable
                                        allUpdated = False
                                        tmpLog.debug('escape since task or dataset is locked')
                                        break
                                    elif missingFileList != []:
                                        # files are missing
                                        tmpErrStr = '{0} files missing in {1}'.format(len(missingFileList),datasetSpec.datasetName)
                                        tmpLog.debug(tmpErrStr)
                                        taskSpec.setErrDiag(tmpErrStr)
                                        allUpdated = False
                                        taskOnHold = True
                                        missingMap[datasetSpec.datasetName] = {'datasetSpec':datasetSpec,
                                                                               'missingFiles':missingFileList}
                                    else:
                                        # reduce the number of files to be read
                                        if 'nFiles' in taskParamMap:
                                            if datasetSpec.isMaster():
                                                taskParamMap['nFiles'] -= nFilesUnique
                                        # reduce the number of files for scout
                                        if useScout:
                                            nChunksForScout = diagMap['nChunksForScout']
                                        # number of master input files
                                        if datasetSpec.isMaster():
                                            checkedMaster = True
                                            nFilesMaster += nFilesUnique
                                    # running task
                                    if diagMap['isRunningTask']:
                                        runningTask = True
                                    # no activated pending input for noWait
                                    if (noWaitParent or inputPreStaging) \
                                            and diagMap['nActivatedPending'] == 0 \
                                            and not (useScout and nChunksForScout <= 0) \
                                            and tmpMetadata['state'] != 'closed' and datasetSpec.isMaster():
                                        tmpErrStr = 'insufficient inputs are ready. '
                                        tmpErrStr += diagMap['errMsg']
                                        tmpLog.debug(tmpErrStr)
                                        taskSpec.setErrDiag(tmpErrStr)
                                        taskOnHold = True
                                        setFrozenTime = False
                                        break
                            tmpLog.debug('end loop')
                    # no mater input
                    if not taskOnHold and not taskBroken and allUpdated and nFilesMaster == 0 and checkedMaster:
                        tmpErrStr = 'no master input files. input dataset is empty'
                        tmpLog.error(tmpErrStr)
                        taskSpec.setErrDiag(tmpErrStr,None)
                        if taskSpec.allowEmptyInput() or noWaitParent:
                            taskOnHold = True
                        else:
                            taskBroken = True
                    # index consistency
                    if not taskOnHold and not taskBroken and len(datasetsIdxConsistency) > 0:
                        self.taskBufferIF.removeFilesIndexInconsistent_JEDI(jediTaskID,datasetsIdxConsistency)
                    # update task status
                    if taskBroken:
                        # task is broken
                        taskSpec.status = 'tobroken'
                        tmpMsg = 'set task_status={0}'.format(taskSpec.status)
                        tmpLog.info(tmpMsg)
                        tmpLog.sendMsg(tmpMsg,self.msgType)
                        allRet = self.taskBufferIF.updateTaskStatusByContFeeder_JEDI(jediTaskID,taskSpec,pid=self.pid)
                    # change task status unless the task is running
                    if not runningTask:
                        # send prestaging request
                        if taskSpec.inputPreStaging() and taskSpec.is_first_contents_feed():
                            tmpStat, tmpErrStr = self.send_prestaging_request(taskSpec, taskParamMap, dsList, tmpLog)
                            if tmpStat:
                                taskSpec.set_first_contents_feed(False)
                            else:
                                tmpLog.debug(tmpErrStr)
                                taskSpec.setErrDiag(tmpErrStr)
                                taskOnHold = True
                        if taskOnHold:
                            # go to pending state
                            if taskSpec.status not in ['broken','tobroken']:
                                taskSpec.setOnHold()
                            tmpMsg = 'set task_status={0}'.format(taskSpec.status)
                            tmpLog.info(tmpMsg)
                            tmpLog.sendMsg(tmpMsg,self.msgType)
                            allRet = self.taskBufferIF.updateTaskStatusByContFeeder_JEDI(jediTaskID,taskSpec,pid=self.pid,setFrozenTime=setFrozenTime)
                        elif allUpdated:
                            # all OK
                            allRet,newTaskStatus = self.taskBufferIF.updateTaskStatusByContFeeder_JEDI(jediTaskID,getTaskStatus=True,pid=self.pid,
                                                                                                       useWorldCloud=taskSpec.useWorldCloud())
                            tmpMsg = 'set task_status={0}'.format(newTaskStatus)
                            tmpLog.info(tmpMsg)
                            tmpLog.sendMsg(tmpMsg,self.msgType)
                        # just unlock
                        retUnlock = self.taskBufferIF.unlockSingleTask_JEDI(jediTaskID,self.pid)
                        tmpLog.debug('unlock not-running task with {0}'.format(retUnlock))
                    else:
                        # just unlock
                        retUnlock = self.taskBufferIF.unlockSingleTask_JEDI(jediTaskID,self.pid)
                        tmpLog.debug('unlock task with {0}'.format(retUnlock))
                    tmpLog.debug('done')
            except Exception as e:
                logger.error('{0} failed in runImpl() with {1}: {2}'.format(self.__class__.__name__, str(e),
                                                                            traceback.format_exc()))


    # update dataset
    def updateDatasetStatus(self,datasetSpec,datasetStatus,tmpLog, datasetState=None):
        # update dataset status
        datasetSpec.status   = datasetStatus
        datasetSpec.lockedBy = None
        if datasetState:
            datasetSpec.state = datasetState
        tmpLog.info('update dataset status to {} state to {}'.format(datasetSpec.status, datasetSpec.state))
        self.taskBufferIF.updateDataset_JEDI(datasetSpec,
                                             {'datasetID':datasetSpec.datasetID,
                                              'jediTaskID':datasetSpec.jediTaskID},
                                             lockTask=True)

    # send prestaging request
    def send_prestaging_request(self, task_spec, task_params_map, ds_list, tmp_log):
        try:
            c = iDDS_Client(idds.common.utils.get_rest_host())
            for datasetSpec in ds_list:
                # get rule
                try:
                    tmp_scope, tmp_name = datasetSpec.datasetName.split(':')
                    tmp_name = re.sub('/$', '', tmp_name)
                except Exception:
                    continue
                if 'prestagingRuleID' in task_params_map:
                    if tmp_name not in task_params_map['prestagingRuleID']:
                        continue
                    rule_id = task_params_map['prestagingRuleID'][tmp_name]
                else:
                    rule_id = self.ddmIF.getInterface(task_spec.vo, task_spec.cloud).getActiveStagingRule(
                        tmp_scope + ':' + tmp_name)
                    if rule_id is None:
                        continue
                # request
                tmp_log.debug('sending request to iDDS for {0}'.format(datasetSpec.datasetName))
                req = {
                    'scope': tmp_scope,
                    'name': tmp_name,
                    'requester': 'panda',
                    'request_type': idds.common.constants.RequestType.StageIn,
                    'transform_tag': idds.common.constants.RequestType.StageIn.value,
                    'status': idds.common.constants.RequestStatus.New,
                    'priority': 0,
                    'lifetime': 30,
                    'request_metadata': {
                    'workload_id': task_spec.jediTaskID,
                    'rule_id': rule_id,
                    },
                }
                tmp_log.debug('req {0}'.format(str(req)))
                ret = c.add_request(**req)
                tmp_log.debug('got requestID={0}'.format(str(ret)))
        except Exception as e:
            return False, 'iDDS error : {0}'.format(str(e))
        return True, None


########## lauch

def launcher(commuChannel,taskBufferIF,ddmIF,vos=None,prodSourceLabels=None):
    p = ContentsFeeder(commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels)
    p.start()



import socket
import errno
import time

import multiprocessing.connection as MP


# set timeout to socket
def SocketClient(address):
    '''
    Return a connection object connected to the socket given by `address`
    '''
    family = MP.address_type(address)
    s = socket.socket( getattr(socket, family) )
    defTimeOut = socket.getdefaulttimeout()
    s.settimeout(30)
    t = MP._init_timeout()

    while 1:
        try:
            s.connect(address)
        except socket.error as e:
            if e.args[0] != errno.ECONNREFUSED or MP._check_timeout(t):
                MP.debug('failed to connect to address %s', address)
                raise
            time.sleep(0.01)
        else:
            break
    else:
        raise
    s.settimeout(defTimeOut)

    fd = MP.duplicate(s.fileno())
    conn = MP._multiprocessing.Connection(fd)
    s.close()
    return conn


#MP.SocketClient = SocketClient
